import unittest

import numpy as np

from clouddrift.wavelet import (
    _morsehigh,
    morse_amplitude,
    morse_freq,
    morse_logspace_freq,
    morse_properties,
    morse_wavelet,
    morse_wavelet_transform,
    wavelet_transform,
)

if __name__ == "__main__":
    unittest.main()


class morse_wavelet_transform_tests(unittest.TestCase):
    def test_morse_wavelet_transform_real(self):
        length = 1023
        radian_frequency = 2 * np.pi / np.logspace(np.log10(10), np.log10(100), 50)
        x = np.random.random(length)
        wtx = morse_wavelet_transform(x, 3, 10, radian_frequency)
        wavelet, _ = morse_wavelet(length, 3, 10, radian_frequency)
        wtx2 = wavelet_transform(x, wavelet)
        self.assertTrue(np.allclose(wtx, wtx2))

    def test_morse_wavelet_transform_complex(self):
        length = 1024
        radian_frequency = 2 * np.pi / np.logspace(np.log10(10), np.log10(100), 50)
        x = np.random.random(length) + 1j * np.random.random(length)
        wtx_p, wtx_n = morse_wavelet_transform(x, 3, 10, radian_frequency, complex=True)
        wavelet, _ = morse_wavelet(length, 3, 10, radian_frequency)
        wtx2 = wavelet_transform(x, wavelet)
        wtx3 = wavelet_transform(np.conj(x), wavelet)
        self.assertTrue(np.allclose(wtx_p, 0.5 * wtx2))
        self.assertTrue(np.allclose(wtx_n, 0.5 * wtx3))

    def test_morse_wavelet_transform_rotary_bandpass(self):
        length = 2048
        radian_frequency = 2 * np.pi / np.logspace(np.log10(10), np.log10(100), 50)
        x = np.random.random(length)
        y = np.random.random(length)
        z = x + 1j * y
        wtx = morse_wavelet_transform(x, 3, 10, radian_frequency, complex=False)
        wty = morse_wavelet_transform(y, 3, 10, radian_frequency, complex=False)
        wp = 0.5 * (wtx + 1j * wty)
        wn = 0.5 * (wtx - 1j * wty)
        wp2, _ = morse_wavelet_transform(z, 3, 10, radian_frequency, complex=True)
        wn2, _ = morse_wavelet_transform(
            np.conj(z), 3, 10, radian_frequency, complex=True
        )
        self.assertTrue(np.allclose(wp, wp2))
        self.assertTrue(np.allclose(wn, wn2))

    def test_morse_wavelet_transform_rotary_energy(self):
        length = 1023
        radian_frequency = 2 * np.pi / np.logspace(np.log10(10), np.log10(100), 50)
        x = np.random.random(length)
        y = np.random.random(length)
        z = x + 1j * y
        wtx = morse_wavelet_transform(
            x, 3, 10, radian_frequency, complex=False, normalization="energy"
        )
        wty = morse_wavelet_transform(
            y, 3, 10, radian_frequency, complex=False, normalization="energy"
        )
        wp = (wtx + 1j * wty) / np.sqrt(2)
        wn = (wtx - 1j * wty) / np.sqrt(2)
        wp2, _ = morse_wavelet_transform(
            z, 3, 10, radian_frequency, complex=True, normalization="energy"
        )
        wn2, _ = morse_wavelet_transform(
            np.conj(z), 3, 10, radian_frequency, complex=True, normalization="energy"
        )
        self.assertTrue(np.allclose(wp, wp2))
        self.assertTrue(np.allclose(wn, wn2))

    def test_morse_wavelet_transform_cos(self):
        f = 0.2
        t = np.arange(0, 1000)
        x = np.cos(2 * np.pi * t * f)
        wtx = morse_wavelet_transform(x, 3, 10, 2 * np.pi * np.array([f]))
        self.assertTrue(np.isclose(np.var(x), 0.5 * np.var(wtx), atol=1e-2))

    def test_morse_wavelet_transform_exp(self):
        f = 0.2
        t = np.arange(0, 1024)
        x = np.exp(1j * 2 * np.pi * t * f)
        wtp, wtn = morse_wavelet_transform(
            x, 3, 10, 2 * np.pi * np.array([f]), complex=True
        )
        self.assertTrue(np.isclose(np.var(x), np.var(wtp) + np.var(wtn), atol=1e-2))


class wavelet_transform_tests(unittest.TestCase):
    def test_wavelet_transform_boundary(self):
        length = 2001
        radian_frequency = 2 * np.pi / np.logspace(np.log10(10), np.log10(100), 50)
        wave, wavef = morse_wavelet(
            length, 2, 4, radian_frequency, order=1, normalization="bandpass"
        )
        x = np.random.random(length)
        w1 = wavelet_transform(x - np.mean(x), wave, boundary="mirror")
        w2 = wavelet_transform(x - np.mean(x), wave, boundary="periodic")
        w3 = wavelet_transform(x - np.mean(x), wave, boundary="zeros")
        s = slice(int(length / 4 - 1), int(length / 4 - 1 + length / 2))
        # not sure why the real part only succeeds
        self.assertTrue(np.allclose(np.real(w1[..., s]), np.real(w2[..., s])))
        self.assertTrue(np.allclose(np.real(w1[..., s]), np.real(w3[..., s])))
        self.assertTrue(np.allclose(np.real(w2[..., s]), np.real(w3[..., s])))

    def test_wavelet_transform_complex(self):
        length = 1000
        radian_frequency = 2 * np.pi / np.logspace(np.log10(10), np.log10(100), 50)
        wave, wavef = morse_wavelet(
            length, 2, 4, radian_frequency, order=1, normalization="bandpass"
        )
        x = np.random.random(length)
        y = np.random.random(length)
        wx = wavelet_transform(x, wave, boundary="mirror")
        wy = wavelet_transform(y, wave, boundary="mirror")
        wp = wavelet_transform(
            x + 1j * y,
            0.5 * wave,
            boundary="mirror",
        )
        wn = wavelet_transform(
            x - 1j * y,
            0.5 * wave,
            boundary="mirror",
        )
        wp2 = 0.5 * (wx + 1j * wy)
        wn2 = 0.5 * (wx - 1j * wy)
        self.assertTrue(np.allclose(wp, wp2, atol=1e-6))
        self.assertTrue(np.allclose(wn, wn2, atol=1e-6))

    def test_wavelet_transform_size(self):
        length = 2046
        m = 10
        order = 2
        radian_frequency = 2 * np.pi * np.array([0.1, 0.2, 0.3])
        gamma = 3
        beta = 4
        x = np.random.random((m, m * 2, length))
        wavelet, _ = morse_wavelet(length, gamma, beta, radian_frequency, order=order)
        wtx = wavelet_transform(x, wavelet)
        self.assertTrue(
            np.shape(wtx) == (m, m * 2, order, len(radian_frequency), length)
        )
        x = np.random.random((length, m, m * 2))
        wavelet, _ = morse_wavelet(length, gamma, beta, radian_frequency, order=order)
        wtx = wavelet_transform(x, wavelet, time_axis=0)
        self.assertTrue(
            np.shape(wtx) == (length, m, m * 2, order, len(radian_frequency))
        )
        x = np.random.random((m, length, m * 2))
        wavelet, _ = morse_wavelet(length, gamma, beta, radian_frequency, order=order)
        wtx = wavelet_transform(x, wavelet, time_axis=1)
        self.assertTrue(
            np.shape(wtx) == (m, length, m * 2, order, len(radian_frequency))
        )

    def test_wavelet_transform_size_axis(self):
        length = 1024
        m = 10
        order = 2
        radian_frequency = 2 * np.pi * np.array([0.1, 0.2, 0.3])
        gamma = 3
        beta = 4
        x = np.random.random((length, m, int(m / 2)))
        wave, _ = morse_wavelet(length, gamma, beta, radian_frequency, order=order)
        w = wavelet_transform(x, wave, time_axis=0)
        self.assertTrue(np.shape(w) == (length, m, m / 2, order, len(radian_frequency)))

    def test_wavelet_transform_centered(self):
        J = 10
        ao = np.logspace(np.log10(5), np.log10(40), J) / 100
        x = np.zeros(2**10)
        wavelet, _ = morse_wavelet(len(x), 2, 4, ao, order=1)
        x[2**9] = 1
        y = wavelet_transform(x, wavelet)
        m = np.argmax(np.abs(y), axis=-1)
        self.assertTrue(np.allclose(m, 2**9))

    def test_wavelet_transform_data_real(self):
        t = np.arange(0, 1000)
        dt = np.diff(t[0:2])
        f = 0.2
        omega = dt * 2 * np.pi * f
        a = 1
        x = a * np.cos(2 * np.pi * t * f)
        gamma = 3
        beta = 10
        waveletb, _ = morse_wavelet(
            np.shape(t)[0], gamma, beta, omega, normalization="bandpass"
        )
        wtxb = wavelet_transform(x, waveletb, boundary="mirror")
        self.assertTrue(np.isclose(np.var(wtxb), 2 * np.var(x), rtol=1e-1))

    def test_wavelet_transform_data_complex(self):
        t = np.arange(0, 1000)
        dt = np.diff(t[0:2])
        f = 0.2
        omega = dt * 2 * np.pi * f
        a = 1
        z = a * np.exp(1j * 2 * np.pi * t * f) + a / 2 * np.exp(-1j * 2 * np.pi * t * f)
        gamma = 3
        beta = 10
        waveletb, _ = morse_wavelet(
            np.shape(t)[0], gamma, beta, omega, normalization="bandpass"
        )
        wtzb = wavelet_transform(z, 0.5 * waveletb, boundary="mirror")
        wtzcb = wavelet_transform(np.conj(z), 0.5 * waveletb, boundary="mirror")
        self.assertTrue(np.isclose(np.var(z), np.var(wtzb) + np.var(wtzcb), rtol=1e-1))


class morse_wavelet_tests(unittest.TestCase):
    def test_morse_wavelet_unitenergy(self):
        radian_frequency = 2 * np.pi / np.logspace(np.log10(5), np.log10(40))
        gamma = 2
        beta = 4
        order = 2
        length = 1000
        wave, _ = morse_wavelet(
            length, gamma, beta, radian_frequency, order=order, normalization="energy"
        )
        nrg = np.sum(np.abs(wave) ** 2, axis=-1)
        self.assertTrue(np.allclose(1, nrg, atol=1e-4))


class morse_freq_tests(unittest.TestCase):
    def test_morse_freq_array(self):
        gamma = np.array([[3, 10, 20], [4, 4, 4]])
        beta = np.array([[50, 100, 200], [150, 250, 300]])
        fm, fe, fi = morse_freq(gamma, beta)
        expected_fm = np.array(
            [
                [2.55436477464518, 1.25892541179417, 1.12201845430196],
                [2.47461600191988, 2.81170662595175, 2.94283095638271],
            ]
        )
        expected_fe = np.array(
            [
                [2.55439315237839, 1.25671816756476, 1.12082018344785],
                [2.47358857550119, 2.81100519592177, 2.94221895404181],
            ]
        )
        expected_fi = np.array(
            [
                [2.55447823649861, 1.25450366161231, 1.11960998925271],
                [2.4725685226628, 2.81030677065388, 2.94160913357547],
            ]
        )
        self.assertTrue(np.allclose(fm, expected_fm))
        self.assertTrue(np.allclose(fe, expected_fe))
        self.assertTrue(np.allclose(fi, expected_fi))

    def test_morse_freq_float(self):
        gamma = 3
        beta = 50
        fm, fe, fi = morse_freq(gamma, beta)
        expected_fm = 2.55436477464518
        expected_fe = 2.55439315237839
        expected_fi = 2.55447823649861
        self.assertTrue(np.isclose(fm, expected_fm))
        self.assertTrue(np.isclose(fe, expected_fe))
        self.assertTrue(np.isclose(fi, expected_fi))

    def test_morse_freq_beta_zero(self):
        gamma = 3
        beta = 0
        fm, fe, fi = morse_freq(gamma, beta)
        expected_fm = 0.884997044500518
        expected_fe = 0.401190287437665
        expected_fi = 0.505468088156089
        self.assertTrue(np.isclose(fm, expected_fm))
        self.assertTrue(np.isclose(fe, expected_fe))
        self.assertTrue(np.isclose(fi, expected_fi))


class morse_logspace_freq_tests(unittest.TestCase):
    def test_morse_logspace_freq_high(self):
        # here we are not testing the morse_logspace_freq function
        gamma = np.array([3])
        beta = np.array([4])
        eta = 0.1
        fhigh = _morsehigh(gamma, beta, eta)
        _, waveletfft = morse_wavelet(10000, gamma, beta, fhigh)
        self.assertTrue(
            np.isclose(
                np.abs(0.5 * waveletfft[0, 0, int(10000 / 2) - 1]), eta, atol=1e-3
            )
        )

    def test_morse_logspace_freq_low(self):
        # to write; requires morsebox: Heisenberg time-frequency box for generalized Morse wavelets.
        self.assertTrue(True)

    def test_morse_logspace_freq_values(self):
        fs = morse_logspace_freq(3, 10, 1024)
        self.assertTrue(
            np.allclose(
                fs[[0, -1]], np.array([2.26342969061515, 0.0761392757859202]), atol=1e-5
            )
        )
        fs = morse_logspace_freq(
            3, 10, 1024, highset=(0.3, np.pi), lowset=(5, 0), density=10
        )
        self.assertTrue(
            np.allclose(
                fs[[0, -1]], np.array([2.45100152921832, 0.0759779680679649]), atol=1e-5
            )
        )
        self.assertTrue(np.shape(fs)[0] == 193)


class morse_properties_tests(unittest.TestCase):
    def test_morse_properties(self):
        gamma = 5
        beta = 6
        expected = np.array([5.47722557505166, 0.365148371670111, 2.8])
        width, skew, kurt = morse_properties(gamma, beta)
        self.assertTrue(np.allclose(expected, np.array([width, skew, kurt])))
        gamma = 2
        beta = 4
        expected = np.array([2.82842712474619, -0.353553390593274, 2.625])
        width, skew, kurt = morse_properties(gamma, beta)
        self.assertTrue(np.allclose(expected, np.array([width, skew, kurt])))


class morse_amplitude_tests(unittest.TestCase):
    def test_morse_amplitude_float(self):
        gamma = 3.0
        beta = 5.0
        self.assertTrue(np.isclose(morse_amplitude(gamma, beta), 4.51966469068946))

    def test_morse_amplitude_array(self):
        gamma = np.array([3, 4, 5])
        beta = np.array([3, 5, 7])
        expected_a = np.array([5.43656365691809, 5.28154010330058, 5.06364231419937])
        self.assertTrue(np.allclose(morse_amplitude(gamma, beta), expected_a))

    def test_morse_amplitude_beta_zero(self):
        gamma = np.array([3, 4, 5])
        beta = np.array([0, 0, 0])
        expected_a = np.array([2, 2, 2])
        self.assertTrue(np.allclose(morse_amplitude(gamma, beta), expected_a))

    def test_morse_amplitude_ndarray(self):
        gamma = np.array([[3, 4], [5, 6]])
        beta = np.array([[5.6, 6.5], [7.5, 8.5]])
        expected_a = np.array(
            [[4.03386834889409, 4.61446982215091], [4.87904507028292, 5.03482799479815]]
        )
        self.assertTrue(np.allclose(morse_amplitude(gamma, beta), expected_a))

    def test_morse_amplitude_energy(self):
        gamma = np.array([[3, 4], [5, 6]])
        beta = np.array([[5.6, 6.5], [7.5, 8.5]])
        expected_a = np.array(
            [[6.95583044131426, 9.24984207652964], [10.9133909718769, 12.2799204953579]]
        )
        self.assertTrue(
            np.allclose(
                morse_amplitude(gamma, beta, normalization="energy"), expected_a
            )
        )



import unittest

from clouddrift.adapters import hurdat2


class hurdat2_tests(unittest.TestCase):
    def test_basin_selection(self) -> None:
        with self.subTest("Select Atlantic Basin"):
            download_requests = hurdat2._get_download_requests("atlantic", "some_path")
            url, _, _ = download_requests[0]

            assert len(download_requests) == 1
            assert url == hurdat2._ATLANTIC_BASIN_URL

        with self.subTest("Select Pacific Basin"):
            download_requests = hurdat2._get_download_requests("pacific", "some_path")
            url, _, _ = download_requests[0]

            assert len(download_requests) == 1
            assert url == hurdat2._PACIFIC_BASIN_URL

        with self.subTest("Select Both Option"):
            download_requests = hurdat2._get_download_requests("both", "some_path")
            atlantic_url, _, _ = download_requests[0]
            pacific_url, _, _ = download_requests[1]

            assert len(download_requests) == 2
            assert atlantic_url == hurdat2._ATLANTIC_BASIN_URL
            assert pacific_url == hurdat2._PACIFIC_BASIN_URL



import os
import unittest
from unittest.mock import Mock, patch

from clouddrift.adapters import gdp6h
from tests.adapters.utils import MultiPatcher


class gdp6h_tests(unittest.TestCase):
    drifter_files: list[str]
    response_mock: Mock
    gdp_mock: Mock

    def setUp(self) -> None:
        super().setUp()
        self.drifter_files = [
            "drifter_6h_0.nc",
            "drifter_6h_1.nc",
            "drifter_6h_2.nc",
            "drifter_6h_3.nc",
            "drifter_6h_4.nc",
        ]
        mock_drifter_ids_str = "\n".join(self.drifter_files)
        data_mock = Mock()
        data_mock.decode = Mock(return_value=mock_drifter_ids_str)
        self.response_mock = Mock()
        self.response_mock.read = Mock(return_value=data_mock)
        self.gdp_mock = Mock()
        self.gdp_mock.order_by_date = Mock(side_effect=lambda _, y: y)

    def test_downloads_all_files_returned(self):
        """
        If neither drifter ids or n_random_id (randomized sample size) are provided as arguments,
        download all of the files.
        """
        with MultiPatcher(
            [
                patch(
                    "clouddrift.adapters.gdp6h.urllib.request.urlopen",
                    Mock(return_value=self.response_mock),
                ),
                patch("clouddrift.adapters.gdp6h.download_with_progress", Mock()),
                patch("clouddrift.adapters.gdp6h.os.makedirs", Mock()),
                patch("clouddrift.adapters.gdp6h.gdp", self.gdp_mock),
            ]
        ):
            ret_drifter_ids = gdp6h.download("some-url.com", "../some/path", None, None)
            assert len(ret_drifter_ids) == len(self.drifter_files)

    def test_downloads_subset_of_files_returned(self):
        """
        If n_random_id is provided, exactly `n_random_id` drifter files should be downloaded
        """

        gdp_mock = Mock()
        gdp_mock.gdp_order_by_date = lambda _, y: y
        with MultiPatcher(
            [
                patch(
                    "clouddrift.adapters.gdp6h.urllib.request.urlopen",
                    Mock(return_value=self.response_mock),
                ),
                patch("clouddrift.adapters.gdp6h.download_with_progress", Mock()),
                patch("clouddrift.adapters.gdp6h.os.makedirs", Mock()),
                patch("clouddrift.adapters.gdp6h.gdp", self.gdp_mock),
            ]
        ):
            ret_drifter_ids = gdp6h.download("some-url.com", "../some/path", None, 2)
            assert len(ret_drifter_ids) == 2

    def test_downloads_selected_drifter_id_files(self):
        """
        If drifter_ids are provided, download only those drifter files
        """
        with MultiPatcher(
            [
                patch(
                    "clouddrift.adapters.gdp1h.urllib.request.urlopen",
                    Mock(return_value=self.response_mock),
                ),
                patch("clouddrift.adapters.gdp6h.download_with_progress", Mock()),
                patch("clouddrift.adapters.gdp6h.os.makedirs", Mock()),
                patch("clouddrift.adapters.gdp6h.gdp", self.gdp_mock),
            ]
        ) as mocks:
            drifter_ids = [0, 1, 2]
            ret_drifter_ids = gdp6h.download(
                "some-url.com", "../some/path", drifter_ids, None
            )
            assert len(ret_drifter_ids) == 3
            mocks[1].assert_called_with(
                [
                    (
                        f"some-url.com/netcdf_1_5000/drifter_6h_{did}.nc",
                        os.path.join("../some/path", f"drifter_6h_{did}.nc"),
                    )
                    for did in ret_drifter_ids
                ]
            )



import shutil

import numpy as np

import tests.utils as testutils
from clouddrift.adapters import hurdat2


class hurdat2_integration_tests(testutils.DisableProgressTestCase):
    def test_load_create_ragged_array(self):
        ra = hurdat2.to_raggedarray()
        ds = ra.to_xarray()
        assert "id" in ds.coords
        assert "time" in ds.coords
        assert len(ds.coords["time"]) == len(ra.coords["time"])
        assert len(ds.coords["id"]) == len(ra.coords["id"])

    def test_conversion(self):
        ra = hurdat2.to_raggedarray()
        ra_non_converted = hurdat2.to_raggedarray(convert=False)
        ds = ra.to_xarray()
        ds_non_converted = ra_non_converted.to_xarray()

        assert np.allclose(
            ds["wind_speed"],
            ds_non_converted["wind_speed"] * hurdat2._METERS_IN_NAUTICAL_MILES / 3600,
            equal_nan=True,
        )

        assert np.allclose(
            ds["pressure"], ds_non_converted["pressure"] * 100, equal_nan=True
        )
        assert np.allclose(
            ds["max_sustained_wind_speed_radius"],
            ds_non_converted["max_sustained_wind_speed_radius"]
            * hurdat2._METERS_IN_NAUTICAL_MILES,
            equal_nan=True,
        )

    @classmethod
    def tearDownClass(cls):
        [shutil.rmtree(dir) for dir in [hurdat2._DEFAULT_FILE_PATH]]



import unittest

import numpy as np
from numpy.lib.scimath import sqrt
from scipy.special import iv, kv  # type: ignore

from clouddrift.sphere import EARTH_DAY_SECONDS
from clouddrift.transfer import (
    _rot,
    _xis,
    ivtilde,
    kvtilde,
    wind_transfer,
)

if __name__ == "__main__":
    unittest.main()


class TransferFunctionTestShapes(unittest.TestCase):
    def setUp(self):
        self.omega = (
            2 * np.pi * np.array([-2, -1.5, -0.5, 0, 0.5, 1, 1.5, 2])
        )  # Angular frequency in radians per day
        self.z = np.array([0, 10, 20, 30])  # Depth in meters
        self.cor_freq = 2 * np.pi * 1.5  # Coriolis frequency in radians per day
        self.delta = 10  # Ekman depth in meters
        self.mu = 0  # Madsen depth in meters
        self.bld = 50  # Boundary layer depth in meters
        self.density = 1025  # Seawater density in kg/m^3

    def test_output_size_wind_transfer_no_slip_mu_is_0(self):
        G, dG1, dG2 = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            self.mu,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
        )
        self.assertEqual(G.shape, (len(self.z), len(self.omega)))
        self.assertEqual(dG1.shape, (len(self.z), len(self.omega)))
        self.assertEqual(dG2.shape, (len(self.z), len(self.omega)))

    def test_output_size_wind_transfer_no_slip_mu_not_0(self):
        G, dG1, dG2 = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            5,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
        )
        self.assertEqual(G.shape, (len(self.z), len(self.omega)))
        self.assertEqual(dG1.shape, (0,))
        self.assertEqual(dG2.shape, (0,))

        G, dG1, dG2 = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            5,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
            method="elipot",
        )
        self.assertEqual(G.shape, (len(self.z), len(self.omega)))
        self.assertEqual(dG1.shape, (0,))
        self.assertEqual(dG2.shape, (0,))

    def test_output_size_wind_transfer_free_slip(self):
        G, dG1, dG2 = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            5,
            self.bld,
            boundary_condition="free-slip",
            density=self.density,
        )
        self.assertEqual(G.shape, (len(self.z), len(self.omega)))
        self.assertEqual(dG1.shape, (0,))
        self.assertEqual(dG2.shape, (0,))


class TransferFunctionTestInputs(unittest.TestCase):
    def setUp(self):
        self.omega = 2 * np.pi * np.array([-2, -1.5, -0.5, 0, 0.5, 1, 1.5, 2])
        self.z = np.array([0, 10, 20, 30])
        self.cor_freq = 2 * np.pi * 1.5
        self.delta = 10
        self.mu = 5
        self.bld = 50
        self.density = 1025

    def test_wind_transfer_negative_z(self):
        with self.assertRaises(ValueError):
            wind_transfer(
                self.omega,
                -self.z,
                self.cor_freq,
                self.delta,
                self.mu,
                self.bld,
                density=self.density,
            )

    def test_wind_transfer_negative_delta(self):
        with self.assertRaises(ValueError):
            wind_transfer(
                self.omega,
                self.z,
                self.cor_freq,
                -self.delta,
                self.mu,
                self.bld,
                density=self.density,
            )

    def test_wind_transfer_negative_mu(self):
        with self.assertRaises(ValueError):
            wind_transfer(
                self.omega,
                self.z,
                self.cor_freq,
                self.delta,
                -self.mu,
                self.bld,
                density=self.density,
            )

    def test_wind_transfer_negative_bld(self):
        with self.assertRaises(ValueError):
            wind_transfer(
                self.omega,
                self.z,
                self.cor_freq,
                self.delta,
                self.mu,
                -self.bld,
                density=self.density,
            )

    def test_wind_transfer_negative_density(self):
        with self.assertRaises(ValueError):
            wind_transfer(
                self.omega,
                self.z,
                self.cor_freq,
                self.delta,
                self.mu,
                self.bld,
                density=-self.density,
            )

    def test_wind_transfer_boundary_condition(self):
        with self.assertRaises(ValueError):
            wind_transfer(
                self.omega,
                self.z,
                self.cor_freq,
                self.delta,
                self.mu,
                self.bld,
                boundary_condition="invalid",
                density=self.density,
            )

    def test_wind_transfer_method(self):
        with self.assertRaises(ValueError):
            wind_transfer(
                self.omega,
                self.z,
                self.cor_freq,
                self.delta,
                self.mu,
                self.bld,
                method="invalid",
                density=self.density,
            )


class TransferFunctionSurfaceValues(unittest.TestCase):
    def setUp(self):
        self.omega = (
            2 * np.pi * np.array([-2, -1.5, -0.5, 0, 0.5, 1, 1.5, 2])
        )  # Angular frequency in radians per day
        self.z = np.array([0, 10, 20, 30])  # Depth in meters
        self.cor_freq = 2 * np.pi * 1.5  # Coriolis frequency in radians per day
        self.delta = 10  # Ekman depth in meters
        self.mu = 0  # Madsen depth in meters
        self.bld = 50  # Boundary layer depth in meters
        self.density = 1025  # Seawater density in kg/m^3

    def test_surface_angle_wind_transfer_no_slip_lilly_nh(self):
        G, _, _ = wind_transfer(
            self.omega,
            self.z[0],
            self.cor_freq,
            self.delta,
            self.mu,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
        )
        self.assertTrue(
            np.allclose(
                np.angle(G),
                np.pi / 4 * np.array([1, 0, -1, -1, -1, -1, -1, -1]),
                atol=1e-1,
            )
        )

    def test_surface_angle_wind_transfer_no_slip_elipot_nh(self):
        G, _, _ = wind_transfer(
            self.omega,
            self.z[0],
            self.cor_freq,
            self.delta,
            self.mu,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
            method="elipot",
        )
        # except at inertial frequency which is not defined for elipot case
        self.assertTrue(
            np.allclose(
                np.angle(np.delete(G, 1)),
                np.pi / 4 * np.array([1, -1, -1, -1, -1, -1, -1]),
                atol=1e-1,
            )
        )

    def test_surface_angle_wind_transfer_no_slip_lilly_sh(self):
        G, _, _ = wind_transfer(
            self.omega,
            self.z[0],
            -self.cor_freq,
            self.delta,
            self.mu,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
        )
        self.assertTrue(
            np.allclose(
                np.angle(G),
                np.pi / 4 * np.array([1, 1, 1, 1, 1, 1, 0, -1]),
                atol=1e-1,
            )
        )

    def test_surface_angle_wind_transfer_no_slip_elipot_sh(self):
        G, _, _ = wind_transfer(
            self.omega,
            self.z[0],
            -self.cor_freq,
            self.delta,
            self.mu,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
            method="elipot",
        )
        # except at inertial frequency which is not defined for elipot case
        self.assertTrue(
            np.allclose(
                np.angle(np.delete(G, 6)),
                np.pi / 4 * np.array([1, 1, 1, 1, 1, 1, -1]),
                atol=1e-1,
            )
        )


class TransferFunctionValues(unittest.TestCase):
    # test for values reported in Lilly and Elipot 2021
    def setUp(self):
        self.omega = np.arange(-10, 10 + 0.01, 0.01) * 2 * np.pi
        self.z = np.arange(0, 50, 5)
        self.cor_freq = 2 * np.pi * 1.5
        self.delta = 20
        self.z0 = 20
        self.mu = self.delta**2 / self.z0
        self.bld = 50
        self.density = 1025

    def test_values_finite_bld(self):
        idx = np.abs(self.omega + self.cor_freq).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            self.mu,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                0.1252763 + 0 * 1j,
                0.10296194 + 0 * 1j,
                0.08472979 + 0 * 1j,
                0.06931472 + 0 * 1j,
                0.05596158 + 0 * 1j,
                0.04418328 + 0 * 1j,
                0.03364722 + 0 * 1j,
                0.02411621 + 0 * 1j,
                0.01541507 + 0 * 1j,
                0.0074108 + 0 * 1j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        self.assertTrue(np.allclose(Gp[:, idx], expected_values, atol=1e-8))

    def test_values_infinite_bld(self):
        idx = np.abs(self.omega).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            self.mu,
            np.inf,
            boundary_condition="no-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                0.04850551 - 0.0396564j,
                0.02829943 - 0.03744607j,
                0.01520417 - 0.03297259j,
                0.00668176 - 0.02797499j,
                0.00117548 - 0.02317678j,
                -0.00230945 - 0.01886255j,
                -0.00442761 - 0.01511781j,
                -0.00561905 - 0.01193716j,
                -0.0061842 - 0.00927553j,
                -0.00633045 - 0.00707324j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        self.assertTrue(np.allclose(Gp[:, idx], expected_values, atol=1e-8))

    def test_values_infinite_bld_mu_is_zero(self):
        idx = np.abs(self.omega).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            0,
            np.inf,
            boundary_condition="no-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                0.05 - 0.05j,
                0.02809557 - 0.04736341j,
                0.01207472 - 0.04115335j,
                0.0011821 - 0.03338043j,
                -0.00553969 - 0.0254163j,
                -0.00907736 - 0.0181115j,
                -0.01033938 - 0.01191774j,
                -0.01009828 - 0.00700083j,
                -0.00896897 - 0.00333703j,
                -0.00741087 - 0.00078996j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        self.assertTrue(np.allclose(Gp[:, idx], expected_values, atol=1e-8))

    def test_values_infinite_bld_delta_is_zero(self):
        idx = np.abs(self.omega).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            0,
            self.mu,
            np.inf,
            boundary_condition="no-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                np.nan + np.nan * 1j,
                0.01603955 - 0.07145549j,
                -0.0083329 - 0.04048001j,
                -0.01373573 - 0.02367659j,
                -0.01399476 - 0.01368128j,
                -0.01260713 - 0.00748527j,
                -0.01076378 - 0.0035753j,
                -0.00891647 - 0.00110311j,
                -0.00723577 + 0.00043968j,
                -0.00577549 + 0.00137177j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        self.assertTrue(
            np.allclose(Gp[:, idx], expected_values, atol=1e-8, equal_nan=True)
        )

    def test_values_finite_bld_mu_is_zero(self):
        idx = np.abs(self.omega).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            0,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                0.04916146 - 0.05044871j,
                0.02728561 - 0.04786423j,
                0.01135701 - 0.04180688j,
                0.00063926 - 0.03427553j,
                -0.00579508 - 0.02661964j,
                -0.00889455 - 0.01965344j,
                -0.00952588 - 0.01377343j,
                -0.00842319 - 0.00906617j,
                -0.00617632 - 0.00539995j,
                -0.00324643 - 0.00249871j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        print(Gp[:, idx])
        self.assertTrue(
            np.allclose(Gp[:, idx], expected_values, atol=1e-8, equal_nan=True)
        )

    def test_values_finite_bld_delta_is_zero(self):
        idx = np.abs(self.omega).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            0,
            self.mu,
            self.bld,
            boundary_condition="no-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                np.nan + np.nan * 1j,
                0.0150223 - 0.07199065j,
                -0.00915018 - 0.04153115j,
                -0.01422336 - 0.02519934j,
                -0.01402997 - 0.0156096j,
                -0.0120768 - 0.00973228j,
                -0.00956707 - 0.00603436j,
                -0.00696735 - 0.00364922j,
                -0.00446539 - 0.00205173j,
                -0.00213448 - 0.0009083j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        self.assertTrue(
            np.allclose(Gp[:, idx], expected_values, atol=1e-8, equal_nan=True)
        )

    def test_values_free_slip(self):
        idx = np.abs(self.omega).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            self.mu,
            self.bld,
            boundary_condition="free-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                0.04653128 - 0.0365929j,
                0.02616164 - 0.03449013j,
                0.01266338 - 0.03030143j,
                0.00359447 - 0.02574129j,
                -0.00253419 - 0.02152846j,
                -0.00666497 - 0.01794904j,
                -0.00940876 - 0.01509089j,
                -0.01116722 - 0.01294936j,
                -0.01220577 - 0.01147741j,
                -0.01269919 - 0.01061011j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        self.assertTrue(
            np.allclose(Gp[:, idx], expected_values, atol=1e-8, equal_nan=True)
        )

    def test_values_free_slip_mu_is_zero(self):
        idx = np.abs(self.omega).argmin()
        G, _, _ = wind_transfer(
            self.omega,
            self.z,
            self.cor_freq,
            self.delta,
            0,
            self.bld,
            boundary_condition="free-slip",
            density=self.density,
        )

        expected_values = np.array(
            [
                0.05083587 - 0.04953873j,
                0.02890206 - 0.0468502j,
                0.01278664 - 0.04048806j,
                0.00171537 - 0.03247495j,
                -0.00529895 - 0.02420514j,
                -0.00928086 - 0.01656615j,
                -0.01118 - 0.01006565j,
                -0.01180644 - 0.00494948j,
                -0.01179884 - 0.00130261j,
                -0.01161306 + 0.00087116j,
            ]
        )
        Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
        self.assertTrue(
            np.allclose(Gp[:, idx], expected_values, atol=1e-8, equal_nan=True)
        )

    # def test_values_free_slip_delta_is_zero(self):
    #     idx = np.abs(self.omega).argmin()
    #     G, _, _ = wind_transfer(
    #         self.omega,
    #         self.z,
    #         self.cor_freq,
    #         0,
    #         self.mu,
    #         self.bld,
    #         boundary_condition="free-slip",
    #         density=self.density,
    #     )

    #     expected_values = np.array(
    #         [
    #             np.nan + np.nan * 1j,
    #             0.0150223 - 0.07199065j,
    #             -0.00915018 - 0.04153115j,
    #             -0.01422336 - 0.02519934j,
    #             -0.01402997 - 0.0156096j,
    #             -0.0120768 - 0.00973228j,
    #             -0.00956707 - 0.00603436j,
    #             -0.00696735 - 0.00364922j,
    #             -0.00446539 - 0.00205173j,
    #             -0.00213448 - 0.0009083j,
    #         ]
    #     )
    #     Gp = G * np.abs(self.cor_freq / EARTH_DAY_SECONDS) * self.density
    #     self.assertTrue(
    #         np.allclose(Gp[:, idx], expected_values, atol=1e-8, equal_nan=True)
    #     )


class TransferFunctionTestGradient(unittest.TestCase):
    delta = 10 ** np.arange(-1, 0.05, 3)
    bld = 10 ** np.arange(np.log10(15.15), 5, 0.05)
    [delta_grid, bld_grid] = np.meshgrid(delta, bld)

    def test_gradient_ekman_case(self):
        # Test the gradient of the transfer function, no-slip
        omega = np.array([1e-4])
        z = 15
        cor_freq = 1e-4
        mu = 0
        delta_delta = 1e-6
        delta_bld = 1e-6
        # initialize transfer functions and gradients
        wind_transfer_init = np.zeros((len(self.delta), len(self.bld)), dtype=complex)
        wind_transfer_ddelta_plus = np.zeros(
            (len(self.delta), len(self.bld)), dtype=complex
        )
        wind_transfer_ddelta_minus = np.zeros(
            (len(self.delta), len(self.bld)), dtype=complex
        )
        wind_transfer_dbld_plus = np.zeros(
            (len(self.delta), len(self.bld)), dtype=complex
        )
        wind_transfer_dbld_minus = np.zeros(
            (len(self.delta), len(self.bld)), dtype=complex
        )
        dG_ddelta = np.zeros((len(self.delta), len(self.bld)), dtype=complex)
        dG_dbld = np.zeros((len(self.delta), len(self.bld)), dtype=complex)

        for i in range(len(self.delta)):
            for j in range(len(self.bld)):
                wind_transfer_init[i, j], dG_ddelta[i, j], dG_dbld[i, j] = (
                    wind_transfer(
                        omega=omega,
                        z=z,
                        cor_freq=cor_freq,
                        delta=self.delta[i],
                        mu=mu,
                        bld=self.bld[j],
                    )
                )
                wind_transfer_ddelta_plus[i, j], _, _ = wind_transfer(
                    omega=omega,
                    z=z,
                    cor_freq=cor_freq,
                    delta=self.delta[i] + delta_delta / 2,
                    mu=mu,
                    bld=self.bld[j],
                )
                wind_transfer_ddelta_minus[i, j], _, _ = wind_transfer(
                    omega=omega,
                    z=z,
                    cor_freq=cor_freq,
                    delta=self.delta[i] - delta_delta / 2,
                    mu=mu,
                    bld=self.bld[j],
                )
                wind_transfer_dbld_plus[i, j], _, _ = wind_transfer(
                    omega=omega,
                    z=z,
                    cor_freq=cor_freq,
                    delta=self.delta[i],
                    mu=mu,
                    bld=self.bld[j] + delta_bld / 2,
                )
                wind_transfer_dbld_minus[i, j], _, _ = wind_transfer(
                    omega=omega,
                    z=z,
                    cor_freq=cor_freq,
                    delta=self.delta[i],
                    mu=mu,
                    bld=self.bld[j] - delta_bld / 2,
                )

        dG_ddelta_fd = (
            wind_transfer_ddelta_plus - wind_transfer_ddelta_minus
        ) / delta_delta
        dG_dbld_fd = (wind_transfer_dbld_plus - wind_transfer_dbld_minus) / delta_bld

        bool_indices = dG_ddelta_fd != 0
        # Calculate eps1 using numpy operations
        eps1 = np.max(
            (
                np.abs(dG_ddelta_fd[bool_indices] - dG_ddelta[bool_indices])
                / sqrt(
                    np.abs(dG_ddelta_fd[bool_indices]) ** 2
                    + np.abs(dG_ddelta[bool_indices]) ** 2
                )
            )
        )

        bool_indices = dG_dbld_fd != 0
        # Calculate eps2 using numpy operations
        eps2 = np.max(
            (
                np.abs(dG_dbld_fd[bool_indices] - dG_dbld[bool_indices])
                / sqrt(
                    np.abs(dG_dbld_fd[bool_indices]) ** 2
                    + np.abs(dG_dbld[bool_indices]) ** 2
                )
            )
        )
        self.assertTrue(
            np.log10(eps1) < -4 and np.log10(eps2) < -4,
            "wind_transfer analytic and numerical gradients match for Ekman case",
        )


# class TransferFunctionTestMethods(unittest.TestCase):
#     def setUp(self):
#         self.z = 15.0  # np.arange(0.1, 101, 1)
#         # need to also test bld = np.inf, K0 = 0 and K1 = 0
#         self.bld = [np.inf, 200.0]
#         self.K0 = [0, 1/10, 1/10]
#         self.K1 = [1,0,1]
#         self.cor_freq = 2 * np.pi * 1.5
#         self.delta = sqrt(2 * np.array(self.K0) / self.cor_freq)
#         self.mu = 2 * np.array(self.K1) / self.cor_freq
#         self.omega = 2 * np.pi * np.array([0.5])
#         self.slipstr1 = "no-slip"
#         self.slipstr2 = "free-slip"

#     def test_method_equivalence_no_slip(self):
#         for s in [1, -1]:
#             for i, (delta, mu) in enumerate(zip(self.delta, self.mu)):
#                 for j, bld in enumerate(self.bld):
#                     Ge, _, _ = wind_transfer(
#                         self.omega,
#                         self.z,
#                         self.cor_freq,
#                         delta,
#                         mu,
#                         bld,
#                         method="elipot",
#                         boundary_condition=self.slipstr1,
#                     )
#                     Gl, _, _ = wind_transfer(
#                         self.omega,
#                         self.z,
#                         self.cor_freq,
#                         delta,
#                         mu,
#                         bld,
#                         method="lilly",
#                         boundary_condition=self.slipstr1,
#                     )
#                     #bool_idx = Ge != np.nan and Gl != np.nan
#                     #print(Ge[bool_idx], Gl[bool_idx])
#                     print(Ge, Gl)
#                     #bool1 = np.allclose(Ge[bool_idx], Gl[bool_idx], atol=1e-8)
#                     bool1 = np.allclose(Ge, Gl, atol=1e-8, equal_nan=True)
#                     self.assertTrue(bool1)

#     def test_method_equivalence_free_slip(self):
#         for s in [1, -1]:
#             for i, (delta, mu) in enumerate(zip(self.delta, self.mu)):
#                 for j, bld in enumerate(self.bld):
#                     Ge, _, _ = wind_transfer(
#                         self.omega,
#                         self.z,
#                         self.cor_freq,
#                         delta,
#                         mu,
#                         bld,
#                         method="elipot",
#                         boundary_condition=self.slipstr2,
#                     )
#                     Gl, _, _ = wind_transfer(
#                         self.omega,
#                         self.z,
#                         self.cor_freq,
#                         delta,
#                         mu,
#                         bld,
#                         method="lilly",
#                         boundary_condition=self.slipstr2,
#                     )
#                     #bool_idx = Ge != np.nan and Gl != np.nan
#                     #print(Ge, Gl)
#                     #bool1 = np.allclose(Ge[bool_idx], Gl[bool_idx], atol=1e-8)
#                     bool1 = np.allclose(Ge, Gl, atol=1e-8, equal_nan=True)
#                     self.assertTrue(bool1)


class TestKvTilde(unittest.TestCase):
    def test_kvtilde(self):
        atol = 1e-10
        for s in [1, -1]:
            z = sqrt(s * 1j) * np.arange(15, 100.01, 0.01).reshape(-1, 1)
            bk0 = np.zeros((len(z), 2), dtype=np.complex128)
            bk = np.zeros_like(bk0)

            for i in range(2):
                bk0[:, i] = (np.exp(z) * kv(i, z)).reshape(-1)
                bk[:, i] = (kvtilde(i, z)).reshape(-1)

            if s == 1:
                test_name = "kvTILDE for z with phase of pi/4"
            else:
                test_name = "kvTILDE for z with phase of -pi/4"

            with self.subTest(test_name=test_name):
                self.assertTrue(
                    np.allclose(np.abs((bk0 - bk) / bk), 0, atol=atol),
                    msg=f"Failed: {test_name}",
                )

            for i in range(2):
                bk[:, i] = kvtilde(i, z, 2).reshape(-1)

            if s == 1:
                test_name = "2-term for z with phase of pi/4, order 0 and 1"
            else:
                test_name = "2-term for z with phase of -pi/4, order 0 and 1"

            with self.subTest(test_name=test_name):
                self.assertTrue(
                    np.allclose(np.abs((bk0[:, :2] - bk) / bk), 0, atol=1e-3),
                    msg=f"Failed: {test_name}",
                )

            bk0 = sqrt(np.pi / (2 * z)) * (1 - 1 / (8 * z))
            bk1 = sqrt(np.pi / (2 * z)) * (1 + 3 / (8 * z))

            if s == 1:
                test_name = "2-term vs. analytic for z with phase of pi/4"
            else:
                test_name = "2-term vs. analytic for z with phase of -pi/4"

            with self.subTest(test_name=test_name):
                self.assertTrue(
                    np.allclose(
                        np.abs((np.hstack([bk0, bk1]) - bk) / bk), 0, atol=1e-15
                    ),
                    msg=f"Failed: {test_name}",
                )


class TestIvTilde(unittest.TestCase):
    def test_ivtilde(self):
        atol = 1e-10
        for s in [1, -1]:
            z = sqrt(s * 1j) * np.arange(23.0, 100.0, 0.01).reshape(-1, 1)
            bi0 = np.zeros((len(z), 2), dtype=np.complex128)
            bi = np.zeros_like(bi0)

            for i in range(2):
                bi0[:, i] = (np.exp(-z) * iv(i, z)).reshape(-1)
                bi[:, i] = (ivtilde(i, z)).reshape(-1)

            if s == 1:
                test_name = "ivTILDE for z with phase of pi/4"
            else:
                test_name = "ivTILDE for z with phase of -pi/4"

            with self.subTest(test_name=test_name):
                self.assertTrue(
                    np.allclose(np.abs((bi0 - bi) / bi), 0, atol=atol),
                    msg=f"Failed: {test_name}",
                )

            for i in range(2):
                bi[:, i] = ivtilde(i, z, 2).reshape(-1)

            if s == 1:
                test_name = "2-term for z with phase of pi/4, order 0 and 1"
            else:
                test_name = "2-term for z with phase of -pi/4, order 0 and 1"

            with self.subTest(test_name=test_name):
                self.assertTrue(
                    np.allclose(np.abs((bi0[:, :2] - bi) / bi), 0, atol=1e-3),
                    msg=f"Failed: {test_name}",
                )

            bi0 = sqrt(1 / (2 * np.pi * z)) * (1 + 1 / (8 * z))
            bi1 = sqrt(1 / (2 * np.pi * z)) * (1 - 3 / (8 * z))

            if s == 1:
                test_name = "2-term vs. analytic for z with phase of pi/4"
            else:
                test_name = "2-term vs. analytic for z with phase of -pi/4"

            with self.subTest(test_name=test_name):
                self.assertTrue(
                    np.allclose(
                        np.abs((np.hstack([bi0, bi1]) - bi) / bi), 0, atol=1e-15
                    ),
                    msg=f"Failed: {test_name}",
                )


class TestXis(unittest.TestCase):
    def test_xis_case1(self):
        s = 1.0
        delta = 30.0
        mu = 20.0
        zo = delta**2 / mu
        z = 15.0
        omega = 5.0
        coriolis_frequency = 6.0
        bld = 100.0
        expected_xiz = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(z, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        expected_xih = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(bld, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        expected_xi0 = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
        assert np.allclose(
            _xis(s, zo, delta, z, omega, coriolis_frequency, bld),
            (expected_xiz, expected_xih, expected_xi0),
        )

    def test_xis_case2(self):
        s = -1.0
        delta = 30.0
        mu = 20.0
        zo = delta**2 / mu
        z = np.array([1.0, 2.0, 3.0])
        omega = np.array([0.1, 0.2, 0.3])
        coriolis_frequency = 0.5
        bld = 7.0
        expected_xiz = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(z, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        expected_xih = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(bld, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        expected_xi0 = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
        assert np.allclose(
            _xis(s, zo, delta, z, omega, coriolis_frequency, bld),
            (expected_xiz, expected_xih, expected_xi0),
        )

    def test_xis_case3(self):
        s = 1
        delta = 1.0
        mu = 1.0
        zo = delta**2 / mu
        z = 0.0
        omega = 0.0
        coriolis_frequency = 0.1
        bld = 10.0
        expected_xiz = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(z, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        expected_xih = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt((1 + np.divide(bld, zo)) * np.abs(1 + omega / coriolis_frequency))
        )
        expected_xi0 = (
            2
            * sqrt(2)
            * _rot(s * np.pi / 4)
            * np.divide(zo, delta)
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
        assert np.allclose(
            _xis(s, zo, delta, z, omega, coriolis_frequency, bld),
            (expected_xiz, expected_xih, expected_xi0),
        )


class TestRot(unittest.TestCase):
    def test_rot_positive(self):
        # Test with positive angle
        angle = np.pi / 4
        expected_result = np.exp(1j * angle)
        self.assertTrue(np.isclose(_rot(angle), expected_result))

    def test_rot_negative(self):
        # Test with negative angle
        angle = -np.pi / 3
        expected_result = np.exp(1j * angle)
        self.assertTrue(np.isclose(_rot(angle), expected_result))

    def test_rot_zero(self):
        # Test with zero angle
        angle = 0
        expected_result = np.exp(1j * angle)
        self.assertTrue(np.isclose(_rot(angle), expected_result))

    def test_rot_random(self):
        # Test with random angle
        angle = np.random.uniform(-3 * np.pi, np.pi)
        expected_result = np.exp(1j * angle)
        self.assertTrue(np.isclose(_rot(angle), expected_result))

    def test_rot_empty(self):
        # Test with empty array
        angles = np.array([])
        expected_results = np.array([])
        assert np.allclose(_rot(angles), expected_results)



import os
import unittest
from unittest import TestCase

import awkward as ak  # type: ignore
import numpy as np
import xarray as xr

from clouddrift import RaggedArray

NETCDF_ARCHIVE = "test_archive.nc"
PARQUET_ARCHIVE = "test_archive.parquet"

if __name__ == "__main__":
    unittest.main()


class raggedarray_tests(TestCase):
    @classmethod
    def setUpClass(self):
        """
        Create ragged array and output netCDF and Parquet file
        """
        self.drifter_id = [1, 2, 3]
        self.rowsize = [10, 8, 2]
        self.nb_obs = np.sum(self.rowsize)
        self.nb_traj = len(self.drifter_id)
        self.attrs_global = {
            "title": "test trajectories",
            "history": "version xyz",
        }
        self.variables_coords = [("id", "traj"), ("time", "obs")]

        # append xr.Dataset to a list
        list_ds = []
        for i in range(0, len(self.rowsize)):
            xr_coords = {}
            xr_coords["id"] = (
                ["traj"],
                [self.drifter_id[i]],
                {"long_name": "variable id", "units": "-"},
            )

            xr_coords["time"] = (
                ["obs"],
                np.ones(self.rowsize[i], dtype="int") * self.drifter_id[i],
                {"long_name": "variable time", "units": "-"},
            )

            xr_data = {}
            xr_data["rowsize"] = (
                ["traj"],
                [self.rowsize[i]],
                {"long_name": "variable rowsize", "units": "-"},
            )
            xr_data["temp"] = (
                ["obs"],
                np.random.rand(self.rowsize[i]),
                {"long_name": "variable temp", "units": "-"},
            )

            list_ds.append(
                xr.Dataset(coords=xr_coords, data_vars=xr_data, attrs=self.attrs_global)
            )

        # create test ragged array
        self.name_coords = ["id", "time"]
        self.name_meta = ["rowsize"]
        self.name_data = ["temp"]
        self.name_dims = {"traj": "rows", "obs": "obs"}
        self.coord_dims = {"id": "traj", "time": "obs"}
        self.ra = RaggedArray.from_files(
            [0, 1, 2],
            lambda i: list_ds[i],
            self.name_coords,
            self.name_meta,
            self.name_data,
            self.name_dims,
            lambda i: self.rowsize[i],
        )

        # output archive
        self.ra.to_netcdf(NETCDF_ARCHIVE)
        self.ra.to_parquet(PARQUET_ARCHIVE)

    @classmethod
    def tearDownClass(self):
        """
        Clean up saved archives
        """
        os.remove(NETCDF_ARCHIVE)
        os.remove(PARQUET_ARCHIVE)

    def test_from_awkward(self):
        ra = RaggedArray.from_awkward(
            ak.from_parquet(PARQUET_ARCHIVE),
            self.name_coords,
            self.name_dims,
            self.coord_dims,
        )
        self.compare_awkward_array(ra.to_awkward())

    def test_from_xarray(self):
        ra = RaggedArray.from_xarray(xr.open_dataset(NETCDF_ARCHIVE), "traj")
        self.compare_awkward_array(ra.to_awkward())

    def test_from_xarray_dim_names(self):
        ds = xr.open_dataset("test_archive.nc")

        ra = RaggedArray.from_xarray(
            ds.rename_dims({"traj": "t", "obs": "o"}),
            rows_dim_name="t",
            obs_dim_name="o",
        )
        self.compare_awkward_array(ra.to_awkward())

    def test_length_ragged_arrays(self):
        """
        Validate the size of the ragged array variables
        """
        self.assertEqual(len(self.ra.coords["id"]), self.nb_traj)
        self.assertEqual(len(self.ra.coords["time"]), self.nb_obs)

        self.assertEqual(len(self.ra.metadata["rowsize"]), self.nb_traj)
        self.assertEqual(len(self.ra.data["temp"]), self.nb_obs)

    def test_variable_attrs(self):
        """
        Validate the variable attributes are properly transferred to the ragged array object.
        Note: as part of this test `long_name` is variable but `units` are always "-"
        """
        for var in ["id", "time", "rowsize", "temp"]:
            self.assertEqual(
                self.ra.attrs_variables[var]["long_name"],
                f"variable {var}",
            )
            self.assertEqual(self.ra.attrs_variables[var]["units"], "-")

    def test_global_attrs(self):
        """
        Validate the global attributes are properly transferred to the ragged array object
        """
        self.assertEqual(self.ra.attrs_global, self.attrs_global)

    def compare_awkward_array(self, ds):
        """
        Compare the returned Awkward Array after initializing from a netCDF/Parquet archive
        with the ragged array object
        """
        # dimensions
        self.assertEqual(len(self.ra.data["temp"]), len(ak.flatten(ds.obs["temp"])))

        # coords
        for var in ["time", "id"]:
            if var == "time":
                self.assertTrue(
                    np.allclose(self.ra.coords[var], ak.flatten(ds.obs[var]).to_numpy())
                )
            else:
                self.assertTrue(
                    np.allclose(self.ra.coords[var], ds.obs[var].to_numpy())
                )
            self.assertTrue(
                self.ra.attrs_variables[var] == ds.obs[var].layout.parameters["attrs"]
            )

        # metadata
        self.assertTrue(
            np.allclose(self.ra.metadata["rowsize"], ds["rowsize"].to_numpy())
        )
        self.assertTrue(
            self.ra.attrs_variables["rowsize"]
            == ds["rowsize"].layout.parameters["attrs"]
        )

        # data
        self.assertTrue(
            np.allclose(self.ra.data["temp"], ak.flatten(ds.obs["temp"]).to_numpy())
        )
        self.assertTrue(
            self.ra.attrs_variables["temp"] == ds.obs["temp"].layout.parameters["attrs"]
        )

    def test_netcdf_output(self):
        """
        Validate the netCDF output archive
        """
        ds = RaggedArray.from_netcdf(NETCDF_ARCHIVE, "traj")
        self.compare_awkward_array(ds.to_awkward())

    def test_parquet_output(self):
        """
        Validate the netCDF output archive
        """
        ds = RaggedArray.from_parquet(
            PARQUET_ARCHIVE, self.name_coords, self.name_dims, self.coord_dims
        )
        self.compare_awkward_array(ds.to_awkward())



import concurrent.futures
import logging
import os
from datetime import datetime
from io import BufferedIOBase, BufferedWriter
from typing import Callable, Sequence
from urllib.error import HTTPError, URLError

import requests
from aiohttp.client_exceptions import (
    ClientConnectorDNSError,
    ClientConnectorError,
    ClientError,
)
from tenacity import (
    RetryCallState,
    WrappedFn,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential_jitter,
)
from tqdm import tqdm

_DISABLE_SHOW_PROGRESS = False  # purely to de-noise our test suite output, should never be used/configured outside of that.


def _before_call(rcs: RetryCallState):
    if rcs.attempt_number > 1:
        _logger.warn(
            f"retrying request for (fn, args): {(rcs.fn, rcs.args)}, attempt: {rcs.attempt_number}"
        )


_CHUNK_SIZE = 1_048_576  # 1MiB
_logger = logging.getLogger(__name__)

standard_retry_protocol: Callable[[WrappedFn], WrappedFn] = retry(
    retry=retry_if_exception(
        lambda ex: isinstance(
            ex,
            (
                requests.Timeout,
                requests.ConnectionError,
                HTTPError,
                URLError,
                ClientConnectorDNSError,
                ClientError,
                ClientConnectorError,
            ),
        )
    ),
    wait=wait_exponential_jitter(
        initial=1.25, jitter=10
    ),  # ~ 20-25 minutes total time before completely failing
    stop=stop_after_attempt(10),
    before=_before_call,
)


def download_with_progress(
    download_map: Sequence[
        tuple[str, BufferedIOBase | str] | tuple[str, BufferedIOBase | str, float]
    ],
    show_list_progress: bool | None = None,
    desc: str = "Downloading files",
    custom_retry_protocol: Callable[[WrappedFn], WrappedFn] | None = None,
):
    if show_list_progress is None:
        show_list_progress = len(download_map) > 20
    if custom_retry_protocol is None:
        retry_protocol = standard_retry_protocol
    else:
        retry_protocol = custom_retry_protocol  # type: ignore

    executor = concurrent.futures.ThreadPoolExecutor()
    futures: dict[
        concurrent.futures.Future[None],
        tuple[str, BufferedIOBase | str],
    ] = dict()
    bar = None

    for request in download_map:
        if len(request) > 2:
            src, dst, exp_size = request
        else:
            src, dst = request
            exp_size = None

        futures[
            executor.submit(
                retry_protocol(_download_with_progress),
                src,
                dst,
                exp_size,
                not show_list_progress,
            )
        ] = (src, dst)

    try:
        if show_list_progress:
            bar = tqdm(
                desc=desc,
                total=len(futures),
                unit="Files",
                disable=_DISABLE_SHOW_PROGRESS,
            )

        for fut in concurrent.futures.as_completed(futures):
            src, dst = futures[fut]
            ex = fut.exception(0)
            if ex is None:
                _logger.debug(f"Finished download job: ({src}, {dst})")
                if bar is not None:
                    bar.update(1)
            else:
                raise ex
    except Exception as e:
        _logger.error(
            f"Got the following exception: {str(e)}, cancelling all other jobs and cleaning up \
              any created resources."
        )
        for x in futures.keys():
            src, dst = futures[x]

            if not x.done():
                x.cancel()

            if isinstance(dst, (str,)) and os.path.exists(dst):
                os.remove(dst)
        raise e
    finally:
        executor.shutdown(True)
        if bar is not None:
            bar.close()


def _download_with_progress(
    url: str,
    output: str | BufferedIOBase,
    expected_size: float | None,
    show_progress: bool,
):
    if isinstance(output, str) and os.path.exists(output):
        _logger.debug(f"File exists {output} checking for updates...")
        local_last_modified = os.path.getmtime(output)

        # Get last modified time of the remote file
        try:
            res = requests.head(url, timeout=5)
            remote_last_modified_str = res.headers.get("Last-Modified")
            if remote_last_modified_str:
                remote_last_modified = datetime.strptime(
                    remote_last_modified_str,
                    "%a, %d %b %Y %H:%M:%S %Z",
                )

                # compare with local modified time
                if local_last_modified >= remote_last_modified.timestamp():
                    _logger.debug(f"File: {output} is up to date; skip download.")
                    return
            else:
                _logger.warning(
                    "Cannot determine if the file has been updated on the remote source. "
                    + "'Last-Modified' header not present in server response."
                )
        finally:
            if res is not None:
                res.close()

    bar: tqdm | None = None
    resp: requests.Response | None = None
    buffer: BufferedWriter | BufferedIOBase | None = None

    try:
        resp = requests.get(url, timeout=10, stream=True)
        temp_output = f"{output}.part" if isinstance(output, str) else None

        if isinstance(output, str) and temp_output is not None:
            buffer = open(temp_output, "wb")
        else:
            # If not a string, at runtime we expect a BufferedIOBase
            if not isinstance(output, BufferedIOBase):
                raise TypeError(
                    "output must be either a string path or a BufferedIOBase-like object."
                )
            buffer = output

        # Assert so mypy knows buffer can't be None
        assert buffer is not None

        # If server returns a length, use that
        content_length = resp.headers.get("Content-Length")
        if content_length:
            expected_size = float(content_length)

        # Show progress if requested
        if show_progress:
            bar = tqdm(
                desc=url,
                total=expected_size,
                unit="B",
                unit_scale=True,
                unit_divisor=_CHUNK_SIZE,
                nrows=2,
                disable=_DISABLE_SHOW_PROGRESS,
            )

        # Write chunks
        for chunk in resp.iter_content(_CHUNK_SIZE):
            if not chunk:
                break
            buffer.write(chunk)
            if bar is not None:
                bar.update(len(chunk))

        # If a path was passed in, rename temp file
        if isinstance(output, str) and temp_output is not None:
            if not buffer.closed:
                buffer.close()
            if os.path.exists(output):
                os.remove(output)
            os.rename(temp_output, output)

    finally:
        if resp is not None:
            resp.close()
        if buffer is not None and isinstance(output, str):
            buffer.close()
        if bar is not None:
            bar.close()


__all__ = ["download_with_progress"]



"""
This module defines functions used to adapt the MOSAiC sea-ice drift dataset as
a ragged-array dataset.

The dataset is hosted at https://doi.org/10.18739/A2KP7TS83.

Reference: Angela Bliss, Jennifer Hutchings, Philip Anderson, Philipp Anhaus,
Hans Jakob Belter, Jørgen Berge, Vladimir Bessonov, Bin Cheng, Sylvia Cole,
Dave Costa, Finlo Cottier, Christopher J Cox, Pedro R De La Torre, Dmitry V Divine,
Gilbert Emzivat, Ying-Chih Fang, Steven Fons, Michael Gallagher, Maxime Geoffrey,
Mats A Granskog, ... Guangyu Zuo. (2022). Sea ice drift tracks from the Distributed
Network of autonomous buoys deployed during the Multidisciplinary drifting Observatory
for the Study of Arctic Climate (MOSAiC) expedition 2019 - 2021. Arctic Data Center.
doi:10.18739/A2KP7TS83.

Example
-------
>>> from clouddrift.adapters import mosaic
>>> ds = mosaic.to_xarray()
"""

import xml.etree.ElementTree as ET
from datetime import datetime
from io import BytesIO

import numpy as np
import pandas as pd
import requests
import xarray as xr

from clouddrift.adapters.utils import download_with_progress, standard_retry_protocol

MOSAIC_VERSION = "2022"


def get_dataframes() -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get the MOSAiC data (obs dimension in the target Dataset) and metadata
    (traj dimension in the target dataset ) as pandas DataFrames."""
    xml = get_repository_metadata()
    filenames, urls = get_file_urls(xml)
    exclude_patterns = ["site_buoy_summary", "buoy_list"]
    data_filenames = [
        f for f in filenames if not any([s in f for s in exclude_patterns])
    ]
    data_urls = [
        f
        for n, f in enumerate(urls)
        if not any([s in filenames[n] for s in exclude_patterns])
    ]
    sensor_ids = [f.split("_")[-1].rstrip(".csv") for f in data_filenames]
    sensor_list_url = urls[
        filenames.index([f for f in filenames if "buoy_list" in f].pop())
    ]
    sensors = pd.read_csv(sensor_list_url)

    # Sort the urls by the order of sensor IDs in the sensor list
    order_index = {id: n for n, id in enumerate(sensors["Sensor ID"])}
    sorted_indices = sorted(
        range(len(sensor_ids)), key=lambda k: order_index[sensor_ids[k]]
    )
    sorted_data_urls = [data_urls[i] for i in sorted_indices]
    buffers = [BytesIO() for _ in range(len(sorted_data_urls))]
    requests = [(url, buffer) for url, buffer in zip(sorted_data_urls, buffers)]

    download_with_progress(requests, desc="Downloading data")
    [b.seek(0) for b in buffers]
    dfs = [pd.read_csv(b) for b in buffers]
    obs_df = pd.concat(dfs)

    # Use the index of the concatenated DataFrame to determine the count/rowsize
    zero_indices = [n for n, val in enumerate(list(obs_df.index)) if val == 0]
    sensors["rowsize"] = np.diff(zero_indices + [len(obs_df)])

    # Make the time column the index of the DataFrame, which will make it a
    # coordinate in the xarray Dataset.
    obs_df.set_index("datetime", inplace=True)
    sensors.set_index("Sensor ID", inplace=True)

    return obs_df, sensors


def get_file_urls(xml: bytes) -> tuple[list[str], list[str]]:
    """Pass the MOSAiC XML string and return the list of filenames and URLs."""
    filenames = [
        tag.text
        for tag in ET.fromstring(xml).findall("./dataset/dataTable/physical/objectName")
        if tag.text
    ]
    urls = [
        tag.text
        for tag in ET.fromstring(xml).findall(
            "./dataset/dataTable/physical/distribution/online/url"
        )
        if tag.text
    ]
    return filenames, urls


def get_repository_metadata() -> bytes:
    """Get the MOSAiC repository metadata as an XML string.
    Pass this string to other get_* functions to extract the data you need.
    """
    url = "https://arcticdata.io/metacat/d1/mn/v2/object/doi:10.18739/A2KP7TS83"
    r = standard_retry_protocol(lambda: requests.get(url))()
    return r.content


def to_xarray():
    """Return the MOSAiC data as an ragged-array Xarray Dataset."""

    # Download the data and metadata as pandas DataFrames.
    obs_df, traj_df = get_dataframes()

    # Dates and datetimes are strings; convert them to datetime64 instances
    # for compatibility with CloudDrift's analysis functions.
    obs_df.index = pd.to_datetime(obs_df.index)
    for col in [
        "Deployment Date",
        "Deployment Datetime",
        "First Data Datetime",
        "Last Data Datetime",
    ]:
        traj_df[col] = pd.to_datetime(traj_df[col])

    # Merge into an Xarray Dataset and rename the dimensions and variables to
    # follow the CloudDrift convention.
    ds = xr.merge([obs_df.to_xarray(), traj_df.to_xarray()])
    ds = ds.rename_dims({"datetime": "obs", "Sensor ID": "traj"}).rename_vars(
        {"datetime": "time", "Sensor ID": "id"}
    )

    # Set variable attributes
    ds["longitude"].attrs = {
        "long_name": "longitude",
        "standard_name": "longitude",
        "units": "degrees_east",
    }

    ds["latitude"].attrs = {
        "long_name": "latitude",
        "standard_name": "latitude",
        "units": "degrees_north",
    }

    # global attributes
    ds.attrs = {
        "title": "Multidisciplinary drifting Observatory for the Study of Arctic Climate (MOSAiC) expedition 2019 - 2021",
        "history": f"Dataset updated in {MOSAIC_VERSION}",
        "date_created": datetime.now().isoformat(),
        "publisher_name": "NSF Arctic Data Center",
        "publisher_url": "https://arcticdata.io/catalog/view/doi:10.18739/A2KP7TS83",
        "license": "Creative Commons Attribution 4.0 International License (http://creativecommons.org/licenses/by/4.0/)",
    }

    return ds



import unittest

import numpy as np
import xarray as xr

from clouddrift.kinematics import (
    inertial_oscillation_from_position,
    kinetic_energy,
    position_from_velocity,
    residual_position_from_displacement,
    spin,
    velocity_from_position,
)
from clouddrift.raggedarray import RaggedArray
from clouddrift.sphere import EARTH_RADIUS_METERS, coriolis_frequency

if __name__ == "__main__":
    unittest.main()


def sample_ragged_array() -> RaggedArray:
    drifter_id = [1, 3, 2]
    longitude = [[-121, -111, 51, 61, 71], [103, 113], [12, 22, 32, 42]]
    latitude = [[-90, -45, 45, 90, 0], [10, 20], [10, 20, 30, 40]]
    t = [[1, 2, 3, 4, 5], [4, 5], [2, 3, 4, 5]]
    test = [
        [True, True, True, False, False],
        [False, False],
        [True, False, False, False],
    ]
    rowsize = [len(x) for x in longitude]
    ids = [[d] * rowsize[i] for i, d in enumerate(drifter_id)]
    attrs_global = {
        "title": "test trajectories",
        "history": "version xyz",
    }
    variables_coords = ["ids", "time", "lon", "lat"]

    coords = {"lon": longitude, "lat": latitude, "ids": ids, "time": t}
    metadata = {"ID": drifter_id, "rowsize": rowsize}
    data = {"test": test}

    # append xr.Dataset to a list
    list_ds = []
    for i in range(0, len(rowsize)):
        xr_coords = {}
        for var in coords.keys():
            xr_coords[var] = (
                ["obs"],
                coords[var][i],
                {"long_name": f"variable {var}", "units": "-"},
            )

        xr_data = {}
        for var in metadata.keys():
            xr_data[var] = (
                ["traj"],
                [metadata[var][i]],
                {"long_name": f"variable {var}", "units": "-"},
            )

        for var in data.keys():
            xr_data[var] = (
                ["obs"],
                data[var][i],
                {"long_name": f"variable {var}", "units": "-"},
            )

        list_ds.append(
            xr.Dataset(coords=xr_coords, data_vars=xr_data, attrs=attrs_global)
        )

    ra = RaggedArray.from_files(
        [0, 1, 2],
        lambda i: list_ds[i],
        variables_coords,
        ["ID", "rowsize"],
        ["test"],
    )

    return ra


class kinetic_energy_tests(unittest.TestCase):
    def test_values_1d_velocity(self):
        u = np.random.random(100)
        self.assertTrue(np.all(kinetic_energy(u) == u**2 / 2))

    def test_1d_velocity_kwarg(self):
        u = np.random.random(100)
        self.assertTrue(
            np.all(kinetic_energy(u) == kinetic_energy(u, np.zeros_like(u)))
        )

    def test_values_2d_velocity(self):
        u = np.random.random(100)
        v = np.random.random(100)
        self.assertTrue(np.all(kinetic_energy(u, v) == (u**2 + v**2) / 2))


class inertial_oscillation_from_position_tests(unittest.TestCase):
    def setUp(self):
        self.INPUT_SIZE = 1440
        self.longitude = np.linspace(0, 45, self.INPUT_SIZE)
        self.latitude = np.linspace(40, 45, self.INPUT_SIZE)
        self.relative_vorticity = 0.1 * coriolis_frequency(self.latitude)

    def test_result_has_same_size_as_input(self):
        x, y = inertial_oscillation_from_position(
            self.longitude, self.latitude, relative_bandwidth=0.10, time_step=3600
        )
        self.assertTrue(np.all(self.longitude.shape == x.shape))
        self.assertTrue(np.all(self.longitude.shape == y.shape))

    def test_relative_vorticity(self):
        x, y = inertial_oscillation_from_position(
            self.longitude,
            self.latitude,
            relative_bandwidth=0.10,
            time_step=3600,
            relative_vorticity=self.relative_vorticity,
        )
        self.assertTrue(np.all(self.longitude.shape == x.shape))
        self.assertTrue(np.all(self.longitude.shape == y.shape))

    def test_time_step(self):
        x, y = inertial_oscillation_from_position(
            self.longitude,
            self.latitude,
            relative_bandwidth=0.10,
            time_step=60 * 20,
        )
        self.assertTrue(np.all(self.longitude.shape == x.shape))
        self.assertTrue(np.all(self.longitude.shape == y.shape))

    def test_simulation_case(self):
        lat0 = 30
        lon0 = 0
        t = np.arange(0, 60 - 1 / 24, 1 / 24)
        f = coriolis_frequency(lat0)
        uv = 0.5 * np.exp(-1j * (f * t * 86400 + 0.5 * np.pi))
        lon1, lat1 = position_from_velocity(
            np.real(uv),
            np.imag(uv),
            t * 86400,
            lon0,
            lat0,
            integration_scheme="forward",
            coord_system="spherical",
        )
        x_expected, y_expected = position_from_velocity(
            np.real(uv),
            np.imag(uv),
            t * 86400,
            0,
            0,
            integration_scheme="forward",
            coord_system="cartesian",
        )
        x_expected = x_expected - np.mean(x_expected)
        y_expected = y_expected - np.mean(y_expected)
        xhat1, yhat1 = inertial_oscillation_from_position(
            lon1, lat1, relative_bandwidth=0.10, time_step=3600
        )
        xhat1 = xhat1 - np.mean(xhat1)
        yhat1 = yhat1 - np.mean(yhat1)
        xhat2, yhat2 = inertial_oscillation_from_position(
            lon1, lat1, wavelet_duration=10, time_step=3600
        )
        xhat2 = xhat2 - np.mean(xhat2)
        yhat2 = yhat2 - np.mean(yhat2)
        m = 10
        self.assertTrue(
            np.allclose(xhat2[m * 24 : -m * 24], x_expected[m * 24 : -m * 24], atol=20)
        )
        self.assertTrue(
            np.allclose(yhat2[m * 24 : -m * 24], y_expected[m * 24 : -m * 24], atol=20)
        )
        self.assertTrue(
            np.allclose(xhat2[m * 24 : -m * 24], xhat1[m * 24 : -m * 24], atol=20)
        )
        self.assertTrue(
            np.allclose(yhat2[m * 24 : -m * 24], yhat1[m * 24 : -m * 24], atol=20)
        )


class residual_position_from_displacement_tests(unittest.TestCase):
    def setUp(self):
        self.INPUT_SIZE = 100
        self.lon = np.rad2deg(
            np.linspace(-np.pi, np.pi, self.INPUT_SIZE, endpoint=False)
        )
        self.lat = np.linspace(0, 45, self.INPUT_SIZE)
        self.x = np.random.rand(self.INPUT_SIZE)
        self.y = np.random.rand(self.INPUT_SIZE)

    def test_result_has_same_size_as_input(self):
        lon, lat = residual_position_from_displacement(
            self.lon, self.lat, self.x, self.y
        )
        self.assertTrue(np.all(self.lon.shape == lon.shape))
        self.assertTrue(np.all(self.lon.shape == lat.shape))

    def test_simple_case(self):
        lon, lat = residual_position_from_displacement(
            1, 0, 2 * np.pi * EARTH_RADIUS_METERS / 360, 0
        )
        self.assertTrue(np.isclose((np.mod(lon, 360), lat), (0, 0)).all())


class position_from_velocity_tests(unittest.TestCase):
    def setUp(self):
        self.INPUT_SIZE = 100
        self.lon = np.rad2deg(
            np.linspace(-np.pi, np.pi, self.INPUT_SIZE, endpoint=False)
        )
        self.lat = np.linspace(0, 45, self.INPUT_SIZE)
        self.time = np.linspace(0, 1e7, self.INPUT_SIZE)
        self.uf, self.vf = velocity_from_position(
            self.lon, self.lat, self.time, difference_scheme="forward"
        )
        self.ub, self.vb = velocity_from_position(
            self.lon, self.lat, self.time, difference_scheme="backward"
        )
        self.uc, self.vc = velocity_from_position(
            self.lon, self.lat, self.time, difference_scheme="centered"
        )

    def test_result_has_same_size_as_input(self):
        lon, lat = position_from_velocity(
            self.uf,
            self.vf,
            self.time,
            self.lon[0],
            self.lat[0],
            integration_scheme="forward",
        )
        self.assertTrue(np.all(self.uf.shape == lon.shape))
        self.assertTrue(np.all(self.uf.shape == lat.shape))

    def test_velocity_position_roundtrip_forward(self):
        lon, lat = position_from_velocity(
            self.uf,
            self.vf,
            self.time,
            self.lon[0],
            self.lat[0],
            integration_scheme="forward",
        )
        self.assertTrue(np.allclose(lon, self.lon))
        self.assertTrue(np.allclose(lat, self.lat))

    def test_velocity_position_roundtrip_backward(self):
        lon, lat = position_from_velocity(
            self.ub,
            self.vb,
            self.time,
            self.lon[0],
            self.lat[0],
            integration_scheme="backward",
        )
        self.assertTrue(np.allclose(lon, self.lon))
        self.assertTrue(np.allclose(lat, self.lat))

    def test_velocity_position_roundtrip_centered(self):
        lon, lat = position_from_velocity(
            self.uc,
            self.vc,
            self.time,
            self.lon[0],
            self.lat[0],
            integration_scheme="centered",
        )
        # Centered scheme damps the 2dx waves so we need a looser tolerance.
        self.assertTrue(np.allclose(lon, self.lon, atol=1e-2))
        self.assertTrue(np.allclose(lat, self.lat, atol=1e-2))

    def test_works_with_xarray_forward(self):
        lon, lat = position_from_velocity(
            xr.DataArray(data=self.uf),
            xr.DataArray(data=self.vf),
            xr.DataArray(data=self.time),
            self.lon[0],
            self.lat[0],
            integration_scheme="forward",
        )
        self.assertTrue(np.allclose(lon, self.lon))
        self.assertTrue(np.allclose(lat, self.lat))

    def test_works_with_xarray_backward(self):
        lon, lat = position_from_velocity(
            xr.DataArray(data=self.ub),
            xr.DataArray(data=self.vb),
            xr.DataArray(data=self.time),
            self.lon[0],
            self.lat[0],
            integration_scheme="backward",
        )
        self.assertTrue(np.allclose(lon, self.lon))
        self.assertTrue(np.allclose(lat, self.lat))

    def test_works_with_xarray_centered(self):
        lon, lat = position_from_velocity(
            xr.DataArray(data=self.uc),
            xr.DataArray(data=self.vc),
            xr.DataArray(data=self.time),
            self.lon[0],
            self.lat[0],
            integration_scheme="centered",
        )
        self.assertTrue(np.allclose(lon, self.lon, atol=1e-2))
        self.assertTrue(np.allclose(lat, self.lat, atol=1e-2))

    def test_works_with_2d_array(self):
        uf = np.reshape(np.tile(self.uf, 4), (4, self.uf.size))
        vf = np.reshape(np.tile(self.vf, 4), (4, self.vf.size))
        time = np.reshape(np.tile(self.time, 4), (4, self.time.size))
        expected_lon = np.reshape(np.tile(self.lon, 4), (4, self.lon.size))
        expected_lat = np.reshape(np.tile(self.lat, 4), (4, self.lat.size))
        lon, lat = position_from_velocity(
            uf,
            vf,
            time,
            self.lon[0],
            self.lat[0],
            integration_scheme="forward",
        )
        self.assertTrue(np.allclose(lon, expected_lon))
        self.assertTrue(np.allclose(lat, expected_lat))
        self.assertTrue(np.allclose(lon.shape, expected_lon.shape))
        self.assertTrue(np.allclose(lon.shape, expected_lat.shape))

    def test_works_with_2d_uv_1d_time(self):
        uf = np.reshape(np.tile(self.uf, 4), (4, self.uf.size))
        vf = np.reshape(np.tile(self.vf, 4), (4, self.vf.size))
        expected_lon = np.reshape(np.tile(self.lon, 4), (4, self.lon.size))
        expected_lat = np.reshape(np.tile(self.lat, 4), (4, self.lat.size))
        lon, lat = position_from_velocity(
            uf,
            vf,
            self.time,
            self.lon[0],
            self.lat[0],
            integration_scheme="forward",
        )
        self.assertTrue(np.allclose(lon, expected_lon))
        self.assertTrue(np.allclose(lat, expected_lat))
        self.assertTrue(np.allclose(lon.shape, expected_lon.shape))
        self.assertTrue(np.allclose(lon.shape, expected_lat.shape))

    def test_time_axis(self):
        uf = np.reshape(np.tile(self.uf, 6), (2, 3, self.uf.size))
        vf = np.reshape(np.tile(self.vf, 6), (2, 3, self.vf.size))
        time = np.reshape(np.tile(self.time, 6), (2, 3, self.time.size))
        expected_lon = np.reshape(np.tile(self.lon, 6), (2, 3, self.lon.size))
        expected_lat = np.reshape(np.tile(self.lat, 6), (2, 3, self.lat.size))

        for time_axis in [0, 1, 2]:
            # Pass inputs with swapped axes and differentiate along that time
            # axis.
            lon, lat = position_from_velocity(
                np.swapaxes(uf, time_axis, -1),
                np.swapaxes(vf, time_axis, -1),
                np.swapaxes(time, time_axis, -1),
                self.lon[0],
                self.lat[0],
                integration_scheme="forward",
                time_axis=time_axis,
            )

            # Swap axes back to compare with the expected result.
            self.assertTrue(np.allclose(np.swapaxes(lon, time_axis, -1), expected_lon))
            self.assertTrue(np.allclose(np.swapaxes(lat, time_axis, -1), expected_lat))
            self.assertTrue(
                np.all(np.swapaxes(lon, time_axis, -1).shape == expected_lon.shape)
            )
            self.assertTrue(
                np.all(np.swapaxes(lat, time_axis, -1).shape == expected_lat.shape)
            )


class velocity_from_position_tests(unittest.TestCase):
    def setUp(self):
        self.INPUT_SIZE = 100
        self.lon = np.rad2deg(np.linspace(-np.pi, np.pi, self.INPUT_SIZE))
        self.lat = np.zeros(self.lon.shape)
        self.time = np.linspace(0, 1e7, self.INPUT_SIZE)
        self.uf, self.vf = velocity_from_position(self.lon, self.lat, self.time)
        self.ub, self.vb = velocity_from_position(
            self.lon, self.lat, self.time, difference_scheme="backward"
        )
        self.uc, self.vc = velocity_from_position(
            self.lon, self.lat, self.time, difference_scheme="centered"
        )

    def test_result_has_same_size_as_input(self):
        self.assertTrue(np.all(self.uf.shape == self.vf.shape == self.lon.shape))
        self.assertTrue(np.all(self.ub.shape == self.vb.shape == self.lon.shape))
        self.assertTrue(np.all(self.uc.shape == self.vc.shape == self.lon.shape))

    def test_schemes_are_self_consistent(self):
        self.assertTrue(np.all(self.uf[:-1] == self.ub[1:]))
        self.assertTrue(
            np.all(np.isclose((self.uf[1:-1] + self.ub[1:-1]) / 2, self.uc[1:-1]))
        )
        self.assertTrue(self.uc[0] == self.uf[0])
        self.assertTrue(self.uc[-1] == self.ub[-1])

    def test_result_value(self):
        u_expected = 2 * np.pi * EARTH_RADIUS_METERS / 1e7
        self.assertTrue(np.all(np.isclose(self.uf, u_expected)))
        self.assertTrue(np.all(np.isclose(self.ub, u_expected)))
        self.assertTrue(np.all(np.isclose(self.uc, u_expected)))

    def test_works_with_xarray_forward(self):
        lon = xr.DataArray(data=self.lon, coords={"time": self.time})
        lat = xr.DataArray(data=self.lat, coords={"time": self.time})
        time = xr.DataArray(data=self.time, coords={"time": self.time})
        uf, vf = velocity_from_position(lon, lat, time, difference_scheme="forward")
        self.assertTrue(np.all(uf == self.uf))
        self.assertTrue(np.all(vf == self.vf))

    def test_works_with_xarray_backward(self):
        lon = xr.DataArray(data=self.lon, coords={"time": self.time})
        lat = xr.DataArray(data=self.lat, coords={"time": self.time})
        time = xr.DataArray(data=self.time, coords={"time": self.time})
        ub, vb = velocity_from_position(lon, lat, time, difference_scheme="backward")
        self.assertTrue(np.all(ub == self.ub))
        self.assertTrue(np.all(vb == self.vb))

    def test_works_with_xarray_centered(self):
        lon = xr.DataArray(data=self.lon, coords={"time": self.time})
        lat = xr.DataArray(data=self.lat, coords={"time": self.time})
        time = xr.DataArray(data=self.time, coords={"time": self.time})
        uc, vc = velocity_from_position(lon, lat, time, difference_scheme="centered")
        self.assertTrue(np.all(uc == self.uc))
        self.assertTrue(np.all(vc == self.vc))

    def test_works_with_2d_array(self):
        lon = np.reshape(np.tile(self.lon, 4), (4, self.lon.size))
        lat = np.reshape(np.tile(self.lat, 4), (4, self.lat.size))
        time = np.reshape(np.tile(self.time, 4), (4, self.time.size))
        expected_uf = np.reshape(np.tile(self.uf, 4), (4, self.uf.size))
        expected_vf = np.reshape(np.tile(self.vf, 4), (4, self.vf.size))
        uf, vf = velocity_from_position(lon, lat, time)
        self.assertTrue(np.all(uf == expected_uf))
        self.assertTrue(np.all(vf == expected_vf))
        self.assertTrue(np.all(uf.shape == expected_uf.shape))
        self.assertTrue(np.all(vf.shape == expected_vf.shape))

    def test_works_with_3d_array(self):
        lon = np.reshape(np.tile(self.lon, 4), (2, 2, self.lon.size))
        lat = np.reshape(np.tile(self.lat, 4), (2, 2, self.lat.size))
        time = np.reshape(np.tile(self.time, 4), (2, 2, self.time.size))
        expected_uf = np.reshape(np.tile(self.uf, 4), (2, 2, self.uf.size))
        expected_vf = np.reshape(np.tile(self.vf, 4), (2, 2, self.vf.size))
        uf, vf = velocity_from_position(lon, lat, time)
        self.assertTrue(np.all(uf == expected_uf))
        self.assertTrue(np.all(vf == expected_vf))
        self.assertTrue(np.all(uf.shape == expected_uf.shape))
        self.assertTrue(np.all(vf.shape == expected_vf.shape))

    def test_works_with_3d_positions_1d_time(self):
        lon = np.reshape(np.tile(self.lon, 4), (2, 2, self.lon.size))
        lat = np.reshape(np.tile(self.lat, 4), (2, 2, self.lat.size))
        expected_uf = np.reshape(np.tile(self.uf, 4), (2, 2, self.uf.size))
        expected_vf = np.reshape(np.tile(self.vf, 4), (2, 2, self.vf.size))
        uf, vf = velocity_from_position(lon, lat, self.time)
        self.assertTrue(np.all(uf == expected_uf))
        self.assertTrue(np.all(vf == expected_vf))
        self.assertTrue(np.all(uf.shape == expected_uf.shape))
        self.assertTrue(np.all(vf.shape == expected_vf.shape))

    def test_time_axis(self):
        lon = np.reshape(np.tile(self.lon, 6), (2, 3, self.lon.size))
        lat = np.reshape(np.tile(self.lat, 6), (2, 3, self.lat.size))
        time = np.reshape(np.tile(self.time, 6), (2, 3, self.time.size))
        expected_uf = np.reshape(np.tile(self.uf, 6), (2, 3, self.uf.size))
        expected_vf = np.reshape(np.tile(self.vf, 6), (2, 3, self.vf.size))

        for time_axis in [0, 1, 2]:
            # Pass inputs with swapped axes and differentiate along that time
            # axis.
            uf, vf = velocity_from_position(
                np.swapaxes(lon, time_axis, -1),
                np.swapaxes(lat, time_axis, -1),
                np.swapaxes(time, time_axis, -1),
                time_axis=time_axis,
            )

            # Swap axes back to compare with the expected result.
            self.assertTrue(np.all(np.swapaxes(uf, time_axis, -1) == expected_uf))
            self.assertTrue(np.all(np.swapaxes(vf, time_axis, -1) == expected_vf))
            self.assertTrue(
                np.all(np.swapaxes(uf, time_axis, -1).shape == expected_uf.shape)
            )
            self.assertTrue(
                np.all(np.swapaxes(vf, time_axis, -1).shape == expected_uf.shape)
            )


class spin_tests(unittest.TestCase):
    def test_simple_spin(self):
        u = np.array([0.16622662, 0.92834973, 0.83375613, 0.99353467])
        v = np.array([0.91568661, 0.71860701, 0.69412382, 0.78336003])
        time = np.arange(4.0)
        self.assertTrue(
            np.allclose(
                spin(u, v, time),
                np.array([-0.84356843, 0.03282956, -0.0310163, -0.02280464]),
            )
        )

    def test_known_zero_spin(self):
        u = np.ones(10)
        v = np.ones(10)
        time = np.arange(10.0)
        self.assertTrue(np.all(spin(u, v, time) == 0))

    def test_multidim_inputs(self):
        u = np.ones((4, 3))
        v = np.ones((4, 3))
        time = np.arange(3.0)
        self.assertTrue(spin(u, v, time).shape == (4, 3))
        self.assertTrue(np.all(spin(u, v, time) == 0))

    def test_raises(self):
        with self.assertRaises(AttributeError):
            spin(0, 0, 0)
        with self.assertRaises(AttributeError):
            spin([0], [0], [0])
        with self.assertRaises(IndexError):
            spin(np.array([0]), np.array([0]), np.array([0]))
        with self.assertRaises(ValueError):
            spin(np.ones(4), np.ones(4), np.arange(3.0))
        with self.assertRaises(ValueError):
            u = np.random.random((4, 3))
            v = np.random.random((4, 3))
            time = np.arange(4.0)
            spin(u, v, time)
        with self.assertRaises(ValueError):
            spin(np.ones(4), np.ones(4), np.arange(4.0), time_axis=2)



import unittest

import numpy as np
import xarray as xr

from clouddrift.signal import (
    analytic_signal,
    cartesian_to_rotary,
    ellipse_parameters,
    modulated_ellipse_signal,
    rotary_to_cartesian,
)

if __name__ == "__main__":
    unittest.main()


class analytic_signal_tests(unittest.TestCase):
    def test_size(self):
        x = np.random.rand(99)
        z = analytic_signal(x)
        self.assertEqual(np.shape(x), np.shape(z))

    def test_imag(self):
        x = np.random.rand(99) + 1j * np.random.rand(99)
        z = analytic_signal(x)
        self.assertTrue(isinstance(z, tuple))

    def test_real_odd(self):
        x = np.random.rand(99)
        z = analytic_signal(x)
        self.assertTrue(np.allclose(x, z.real))

    def test_real_even(self):
        x = np.random.rand(100)
        z = analytic_signal(x)
        self.assertTrue(np.allclose(x, z.real))

    def test_imag_odd(self):
        z = np.random.rand(99) + 1j * np.random.rand(99)
        wp, wn = analytic_signal(z)
        self.assertTrue(np.allclose(z, wp + np.conj(wn)))

    def test_imag_even(self):
        z = np.random.rand(100) + 1j * np.random.rand(100)
        wp, wn = analytic_signal(z)
        self.assertTrue(np.allclose(z, wp + np.conj(wn)))

    def test_boundary(self):
        x = np.random.rand(99)
        z1 = analytic_signal(x, boundary="mirror")
        z2 = analytic_signal(x, boundary="zeros")
        z3 = analytic_signal(x, boundary="periodic")
        self.assertTrue(np.allclose(x, z1.real))
        self.assertTrue(np.allclose(x, z2.real))
        self.assertTrue(np.allclose(x, z3.real))

    def test_ndarray(self):
        x = np.random.random((9, 11, 13))
        for n in range(3):
            z = analytic_signal(x, time_axis=n)
            self.assertTrue(np.allclose(x, z.real))

    def test_xarray(self):
        x = xr.DataArray(data=np.random.random((9, 11, 13)))
        for n in range(3):
            z = analytic_signal(x, time_axis=n)
            self.assertTrue(np.allclose(x, z.real))


class cartesian_to_rotary_tests(unittest.TestCase):
    def test_size(self):
        ua = np.random.rand(99) + 1j * np.random.rand(99)
        va = np.random.rand(99) + 1j * np.random.rand(99)
        wp, wn = cartesian_to_rotary(ua, va)
        self.assertEqual(np.shape(ua), np.shape(wp))
        self.assertEqual(np.shape(ua), np.shape(wn))

    def test_array(self):
        ua = np.random.random(99) + 1j * np.random.random(99)
        va = np.random.random(99) + 1j * np.random.random(99)
        wp, wn = cartesian_to_rotary(ua, va)
        self.assertTrue(np.allclose(np.real(ua) + 1j * np.real(va), wp + np.conj(wn)))

    def test_ndarray(self):
        ua = np.random.rand(99, 10, 101) + 1j * np.random.rand(99, 10, 101)
        va = np.random.rand(99, 10, 101) + 1j * np.random.rand(99, 10, 101)
        wp, wn = cartesian_to_rotary(ua, va)
        self.assertTrue(np.allclose(np.real(ua) + 1j * np.real(va), wp + np.conj(wn)))

    def test_xarray(self):
        ua = xr.DataArray(data=np.random.rand(99, 100) + 1j * np.random.rand(99, 100))
        va = xr.DataArray(data=np.random.rand(99, 100) + 1j * np.random.rand(99, 100))
        wp, wn = cartesian_to_rotary(ua, va)
        self.assertTrue(np.allclose(np.real(ua) + 1j * np.real(va), wp + np.conj(wn)))

    def test_invert_cartesian_to_rotary(self):
        u = np.random.rand(99)
        v = np.random.rand(99)
        ua = analytic_signal(u)
        va = analytic_signal(v)
        wp, wn = cartesian_to_rotary(ua, va)
        ua_, va_ = rotary_to_cartesian(wp, wn)
        self.assertTrue(np.allclose(ua, ua_))
        self.assertTrue(np.allclose(va, va_))
        self.assertTrue(np.allclose(u, np.real(ua_)))
        self.assertTrue(np.allclose(v, np.real(va_)))


class ellipse_parameters_tests(unittest.TestCase):
    def setUp(self):
        t = np.arange(0, 926)
        self.kappa = 3 * np.exp(2 * 0.393 * (t / 1000 - 1))
        self.lamb = 0.5 * np.ones_like(t)
        self.phi = np.pi * 2 * t / 1000 * 5
        self.theta = np.pi / 4 + self.phi / 14.45
        self.a = self.kappa * np.sqrt(1 + np.abs(self.lamb))
        self.b = np.sign(self.lamb) * self.kappa * np.sqrt(1 - np.abs(self.lamb))
        self.xa = np.exp(1j * self.phi) * (
            self.a * np.cos(self.theta) + 1j * self.b * np.sin(self.theta)
        )
        self.ya = np.exp(1j * self.phi) * (
            self.a * np.sin(self.theta) - 1j * self.b * np.cos(self.theta)
        )

    def test_result_has_same_size_as_input(self):
        kappa, lamb, theta, phi = ellipse_parameters(self.xa, self.ya)
        self.assertEqual(np.shape(kappa), np.shape(self.xa))
        self.assertEqual(np.shape(lamb), np.shape(self.xa))
        self.assertEqual(np.shape(theta), np.shape(self.xa))
        self.assertEqual(np.shape(phi), np.shape(self.xa))

    def test_xarray_size(self):
        xa = xr.DataArray(data=self.xa)
        ya = xr.DataArray(data=self.ya)
        kappa, lamb, theta, phi = ellipse_parameters(xa, ya)
        self.assertEqual(np.shape(kappa), np.shape(xa))
        self.assertEqual(np.shape(lamb), np.shape(xa))
        self.assertEqual(np.shape(theta), np.shape(xa))
        self.assertEqual(np.shape(phi), np.shape(xa))

    def test_result_is_correct(self):
        kappa, lamb, theta, phi = ellipse_parameters(self.xa, self.ya)
        self.assertTrue(np.allclose(kappa, self.kappa, atol=1e-2))
        self.assertTrue(np.allclose(lamb, self.lamb, atol=1e-2))
        self.assertTrue(
            np.allclose(
                np.mod(theta + phi, np.pi),
                np.mod(self.theta + self.phi, np.pi),
                atol=1e-2,
            )
        )

    def test_invert_ellipse_parameters(self):
        kappa, lamb, theta, phi = ellipse_parameters(self.xa, self.ya)
        xa, ya = modulated_ellipse_signal(kappa, lamb, theta, phi)
        self.assertTrue(np.allclose(np.abs(self.xa), np.abs(xa), atol=1e-2))
        self.assertTrue(np.allclose(np.abs(self.ya), np.abs(ya), atol=1e-2))
        self.assertTrue(
            np.allclose(
                np.mod(np.angle(self.xa), np.pi), np.mod(np.angle(xa), np.pi), atol=1e-2
            )
        )
        self.assertTrue(
            np.allclose(
                np.mod(np.angle(self.ya), np.pi), np.mod(np.angle(ya), np.pi), atol=1e-2
            )
        )


class modulated_ellipse_signal_tests(unittest.TestCase):
    def setUp(self):
        self.phi = np.linspace(0, 10 * 2 * np.pi, 1000) + np.pi / 3
        self.theta = -np.pi / 4 * np.ones_like(self.phi)
        self.a = 2 * np.ones_like(self.phi)
        self.b = -1 * np.ones_like(self.phi)
        self.kappa = np.sqrt(0.5 * (self.a**2 + self.b**2)) * np.ones_like(self.phi)
        self.lamb = (
            np.ones_like(self.phi)
            * np.sign(self.b)
            * (self.a**2 - self.b**2)
            / (self.a**2 + self.b**2)
        )
        self.xa = np.exp(1j * self.phi) * (
            self.a * np.cos(self.theta) + 1j * self.b * np.sin(self.theta)
        )
        self.ya = np.exp(1j * self.phi) * (
            self.a * np.sin(self.theta) - 1j * self.b * np.cos(self.theta)
        )

    def test_result_has_same_size_as_input(self):
        xa, ya = modulated_ellipse_signal(self.kappa, self.lamb, self.theta, self.phi)
        self.assertEqual(np.shape(xa), np.shape(self.kappa))
        self.assertEqual(np.shape(ya), np.shape(self.kappa))

    def test_xarray_size(self):
        kappa = xr.DataArray(data=self.kappa)
        lamb = xr.DataArray(data=self.lamb)
        theta = xr.DataArray(data=self.theta)
        phi = xr.DataArray(data=self.phi)
        xa, ya = modulated_ellipse_signal(kappa, lamb, theta, phi)
        self.assertEqual(np.shape(xa), np.shape(kappa))
        self.assertEqual(np.shape(ya), np.shape(kappa))

    def test_result_is_correct(self):
        xa, ya = modulated_ellipse_signal(self.kappa, self.lamb, self.theta, self.phi)
        self.assertTrue(np.allclose(xa, self.xa, atol=1e-2))
        self.assertTrue(np.allclose(ya, self.ya, atol=1e-2))

    def test_invert_modulated_ellipse_signal(self):
        xa, ya = modulated_ellipse_signal(self.kappa, self.lamb, self.theta, self.phi)
        kappa, lamb, theta, phi = ellipse_parameters(xa, ya)
        self.assertTrue(np.allclose(kappa, self.kappa, atol=1e-2))
        self.assertTrue(np.allclose(lamb, self.lamb, atol=1e-2))
        self.assertTrue(
            np.allclose(
                np.mod(phi + theta, np.pi),
                np.mod(self.phi + self.theta, np.pi),
                atol=1e-2,
            )
        )


class rotary_to_cartesian_tests(unittest.TestCase):
    def test_size(self):
        wp = np.random.rand(99) + 1j * np.random.rand(99)
        wn = np.random.rand(99) + 1j * np.random.rand(99)
        ua, va = rotary_to_cartesian(wp, wn)
        self.assertEqual(np.shape(ua), np.shape(wp))
        self.assertEqual(np.shape(va), np.shape(wp))

    def test_array(self):
        wp = np.random.random(100) + 1j * np.random.random(100)
        wn = np.random.random(100) + 1j * np.random.random(100)
        ua, va = rotary_to_cartesian(wp, wn)
        self.assertTrue(np.allclose(np.real(ua) + 1j * np.real(va), wp + np.conj(wn)))

    def test_ndarray(self):
        wp = np.random.rand(99, 10, 101) + 1j * np.random.rand(99, 10, 101)
        wn = np.random.rand(99, 10, 101) + 1j * np.random.rand(99, 10, 101)
        ua, va = rotary_to_cartesian(wp, wn)
        self.assertTrue(np.allclose(np.real(ua) + 1j * np.real(va), wp + np.conj(wn)))

    def test_xarray(self):
        wp = xr.DataArray(data=np.random.rand(99)) + 1j * xr.DataArray(
            data=np.random.rand(99)
        )
        wn = xr.DataArray(data=np.random.rand(99)) + 1j * xr.DataArray(
            data=np.random.rand(99)
        )
        ua, va = rotary_to_cartesian(wp, wn)
        self.assertTrue(np.allclose(np.real(ua) + 1j * np.real(va), wp + np.conj(wn)))

    def test_invert_rotary_to_cartesian(self):
        wp = np.random.random(100) + 1j * np.random.random(100)
        wn = np.random.random(100) + 1j * np.random.random(100)
        ua, va = rotary_to_cartesian(wp, wn)
        wp_, wn_ = cartesian_to_rotary(ua, va)
        self.assertTrue(np.allclose(wp, wp_))
        self.assertTrue(np.allclose(wn, wn_))



"""
This module provides functions for computing wavelet transforms and time-frequency analyses,
notably using generalized Morse wavelets.

The Python code in this module was translated from the MATLAB implementation
by J. M. Lilly in the jWavelet module of jLab (http://jmlilly.net/code.html).

Lilly, J. M. (2021), jLab: A data analysis package for Matlab, v.1.7.1,
doi:10.5281/zenodo.4547006, http://www.jmlilly.net/software.

jLab is licensed under the Creative Commons Attribution-Noncommercial-ShareAlike
License (https://creativecommons.org/licenses/by-nc-sa/4.0/). The code that is
directly translated from jLab/jWavelet is licensed under the same license.
Any other code that is added to this module and that is specific to Python and
not the MATLAB implementation is licensed under CloudDrift's MIT license.
"""

import numpy as np
from scipy.special import gamma as _gamma
from scipy.special import gammaln as _lgamma


def morse_wavelet_transform(
    x: np.ndarray,
    gamma: float,
    beta: float,
    radian_frequency: np.ndarray,
    complex: bool = False,
    order: int = 1,
    normalization: str = "bandpass",
    boundary: str = "periodic",
    time_axis: int = -1,
) -> tuple[np.ndarray, np.ndarray] | np.ndarray:
    """
    Apply a continuous wavelet transform to an input signal using the generalized Morse
    wavelets of Olhede and Walden (2002). The wavelet transform is normalized differently
    for complex-valued input than for real-valued input, and this in turns depends on whether the
    optional argument ``normalization`` is set to ``"bandpass"`` or ``"energy"`` normalizations.

    Parameters
    ----------
    x : np.ndarray
        Real- or complex-valued signals. The time axis is assumed to be the last. If not, specify optional
        argument `time_axis`.
    gamma : float
       Gamma parameter of the Morse wavelets.
    beta : float
       Beta parameter of the Morse wavelets.
    radian_frequency : np.ndarray
       An array of radian frequencies at which the Fourier transform of the wavelets
       reach their maximum amplitudes. ``radian_frequency`` is typically between 0 and 2 * np.pi * 0.5,
       the normalized Nyquist radian frequency.
    complex : boolean, optional
        Specify explicitely if the input signal ``x`` is a complex signal. Default is False which
        means that the input is real but that is not explicitely tested by the function.
        This choice affects the normalization of the outputs and their interpretation.
        See examples below.
    time_axis : int, optional
        Axis on which the time is defined for input ``x`` (default is last, or -1).
    normalization : str, optional
        Normalization for the wavelet transforms. By default it is assumed to be
        ``"bandpass"`` which uses a bandpass normalization, meaning that the FFT
        of the wavelets have peak value of 2 for all central frequencies
        ``radian_frequency``. However, if the optional argument ``complex=True``
        is specified, the wavelets will be divided by 2 so that the total
        variance of the input complex signal is equal to the sum of the
        variances of the returned analytic (positive) and conjugate analytic
        (negative) parts. See examples below. The other option is ``"energy"``
        which uses the unit energy normalization. In this last case, the
        time-domain wavelet energies ``np.sum(np.abs(wave)**2)`` are always
        unity.
    boundary : str, optional
        The boundary condition to be imposed at the edges of the input signal ``x``.
        Allowed values are ``"mirror"``, ``"zeros"``, and ``"periodic"``. Default is ``"periodic"``.
    order : int, optional
        Order of Morse wavelets, default is 1.

    Returns
    -------
    If the input signal is real as specificied by ``complex=False``:

    wtx : np.ndarray
        Time-domain wavelet transform of input ``x`` with shape ((x shape without time_axis), orders, frequencies, time_axis)
        but with dimensions of length 1 removed (squeezed).

    If the input signal is complex as specificied by ``complex=True``, a tuple is returned:

    wtx_p : np.array
        Time-domain positive wavelet transform of input ``x`` with shape ((x shape without time_axis), frequencies, orders),
        but with dimensions of length 1 removed (squeezed).
    wtx_n : np.array
        Time-domain negative wavelet transform of input ``x`` with shape ((x shape without time_axis), frequencies, orders),
        but with dimensions of length 1 removed (squeezed).

    Examples
    --------
    Apply a wavelet transform with a Morse wavelet with gamma parameter 3, beta parameter 4,
    at radian frequency 0.2 cycles per unit time:

    >>> x = np.random.random(1024)
    >>> wtx = morse_wavelet_transform(x, 3, 4, np.array([2*np.pi*0.2]))

    Apply a wavelet transform with a Morse wavelet with gamma parameter 3, beta parameter 4,
    for a complex input signal at radian frequency 0.2 cycles per unit time. This case returns the
    analytic and conjugate analytic components:

    >>> z = np.random.random(1024) + 1j*np.random.random(1024)
    >>> wtz_p, wtz_n = morse_wavelet_transform(z, 3, 4, np.array([2*np.pi*0.2]), complex=True)

    The same result as above can be otained by applying the Morse transform on the real and imaginary
    component of z and recombining the results as follows for the "bandpass" normalization:
    >>> wtz_real = morse_wavelet_transform(np.real(z)), 3, 4, np.array([2*np.pi*0.2]))
    >>> wtz_imag = morse_wavelet_transform(np.imag(z)), 3, 4, np.array([2*np.pi*0.2]))
    >>> wtz_p, wtz_n = (wtz_real + 1j*wtz_imag) / 2, (wtz_real - 1j*wtz_imag) / 2

    For the "energy" normalization, the analytic and conjugate analytic components are obtained as follows
    with this alternative method:
    >>> wtz_real = morse_wavelet_transform(np.real(z)), 3, 4, np.array([2*np.pi*0.2]))
    >>> wtz_imag = morse_wavelet_transform(np.imag(z)), 3, 4, np.array([2*np.pi*0.2]))
    >>> wtz_p, wtz_n = (wtz_real + 1j*wtz_imag) / np.sqrt(2), (wtz_real - 1j*wtz_imag) / np.sqrt(2)

    The input signal can have an arbitrary number of dimensions but its ``time_axis`` must be
    specified if it is not the last:

    >>> x = np.random.random((1024,10,15))
    >>> wtx = morse_wavelet_transform(x, 3, 4, np.array([2*np.pi*0.2]), time_axis=0)

    The default way to handle the boundary conditions is to assume periodicity of the
    input but this can be changed by specifying another boundary condition:

    >>> x = np.random.random((10,15,1024))
    >>> wtx = morse_wavelet_transform(x, 3, 4, np.array([2*np.pi*0.2]), boundary="mirror")

    This function can be used to conduct a time-frequency analysis of the input signal by specifying
    a range of randian frequencies using the ``morse_logspace_freq`` function as an example:

    >>> x = np.random.random(1024)
    >>> gamma = 3
    >>> beta = 4
    >>> radian_frequency = morse_logspace_freq(gamma, beta, np.shape(x)[0])
    >>> wtx = morse_wavelet_transform(x, gamma, beta, radian_frequency)

    Raises
    ------
    ValueError
        If the time axis is outside of the valid range ([-1, np.ndim(x)-1]).
        If boundary optional argument is not in ["mirror", "zeros", "periodic"]``.
        If normalization optional argument is not in ["bandpass", "energy"]``.

    See Also
    --------
    :func:`morse_wavelet`, :func:`wavelet_transform`, :func:`morse_logspace_freq`

    """
    # time_axis must be in valid range
    if time_axis < -1 or time_axis > len(x.shape) - 1:
        raise ValueError(
            f"time_axis ({time_axis}) is outside of the valid range ([-1,"
            f" {len(x.shape) - 1}])."
        )
    # generate the wavelet
    wavelet, _ = morse_wavelet(
        np.shape(x)[time_axis],
        gamma,
        beta,
        radian_frequency,
        normalization=normalization,
        order=order,
    )

    # apply the wavelet transform, distinguish complex and real cases
    if complex:
        # imaginary case, divide by 2 the wavelet and return analytic and conjugate analytic
        if normalization == "bandpass":
            wtx_p = wavelet_transform(
                0.5 * x, wavelet, boundary=boundary, time_axis=time_axis
            )
            wtx_n = wavelet_transform(
                np.conj(0.5 * x), wavelet, boundary=boundary, time_axis=time_axis
            )
        elif normalization == "energy":
            wtx_p = wavelet_transform(
                x / np.sqrt(2), wavelet, boundary=boundary, time_axis=time_axis
            )
            wtx_n = wavelet_transform(
                np.conj(x / np.sqrt(2)), wavelet, boundary=boundary, time_axis=time_axis
            )
        wtx = wtx_p, wtx_n

    elif not complex:
        # real case
        wtx = wavelet_transform(x, wavelet, boundary=boundary, time_axis=time_axis)

    else:
        raise ValueError(
            "`complex` optional argument must be boolean 'True' or 'False'"
        )

    return wtx


def wavelet_transform(
    x: np.ndarray,
    wavelet: np.ndarray,
    boundary: str = "periodic",
    time_axis: int = -1,
    freq_axis: int = -2,
    order_axis: int = -3,
) -> np.ndarray:
    """
    Apply a continuous wavelet transform to an input signal using an input wavelet
    function. Such wavelet can be provided by the function ``morse_wavelet``.

    Parameters
    ----------
    x : np.ndarray
        Real- or complex-valued signals.
    wavelet : np.ndarray
        A suite of time-domain wavelets, typically returned by the function ``morse_wavelet``.
        The length of the time axis of the wavelets must be the last one and matches the
        length of the time axis of x. The other dimensions (axes) of the wavelets (such as orders and frequencies) are
        typically organized as orders, frequencies, and time, unless specified by optional arguments freq_axis and order_axis.
        The normalization of the wavelets is assumed to be "bandpass", if not, use kwarg normalization="energy", see ``morse_wavelet``.
    boundary : str, optional
        The boundary condition to be imposed at the edges of the input signal ``x``.
        Allowed values are ``"mirror"``, ``"zeros"``, and ``"periodic"``. Default is ``"periodic"``.
    time_axis : int, optional
        Axis on which the time is defined for input ``x`` (default is last, or -1). Note that the time axis of the
        wavelets must be last.
    freq_axis : int, optional
        Axis of ``wavelet`` for the frequencies (default is second or 1)
    order_axis : int, optional
        Axis of ``wavelet`` for the orders (default is first or 0)

    Returns
    -------
    wtx : np.ndarray
        Time-domain wavelet transform of ``x`` with shape ((x shape without time_axis), orders, frequencies, time_axis)
        but with dimensions of length 1 removed (squeezed).

    Examples
    --------
    Apply a wavelet transform with a Morse wavelet with gamma parameter 3, beta
    parameter 4, at radian frequency 0.2 cycles per unit time:

    >>> x = np.random.random(1024)
    >>> wavelet, _ = morse_wavelet(1024, 3, 4, np.array([2*np.pi*0.2]))
    >>> wtx = wavelet_transform(x, wavelet)

    The input signal can have an arbitrary number of dimensions but its
    ``time_axis`` must be specified if it is not the last:

    >>> x = np.random.random((1024,10,15))
    >>> wavelet, _ = morse_wavelet(1024, 3, 4, np.array([2*np.pi*0.2]))
    >>> wtx = wavelet_transform(x, wavelet,time_axis=0)

    Raises
    ------
    ValueError
        If the time axis is outside of the valid range ([-1, N-1]).
        If the shape of time axis is different for input signal and wavelet.
        If boundary optional argument is not in ["mirror", "zeros", "periodic"]``.

    See Also
    --------
    :func:`morse_wavelet`, :func:`morse_wavelet_transform`, :func:`morse_freq`
    """
    # time_axis must be in valid range
    if time_axis < -1 or time_axis > len(x.shape) - 1:
        raise ValueError(
            f"time_axis ({time_axis}) is outside of the valid range ([-1,"
            f" {len(x.shape) - 1}])."
        )
    # Positions and time arrays must have the same shape.
    if x.shape[time_axis] != wavelet.shape[-1]:
        raise ValueError("x and wavelet time axes must have the same length.")

    wavelet_ = np.moveaxis(wavelet, [freq_axis, order_axis], [-2, -3])

    # if x is of dimension 1 we need to expand
    # otherwise make sure time axis is last
    if np.ndim(x) < 2:
        x_ = np.expand_dims(x, axis=0)
    else:
        x_ = np.moveaxis(x, time_axis, -1)

    # add detrending option eventually

    # apply boundary conditions
    if boundary == "mirror":
        x_ = np.concatenate((np.flip(x_, axis=-1), x_, np.flip(x_, axis=-1)), axis=-1)
    elif boundary == "zeros":
        x_ = np.concatenate((np.zeros_like(x_), x_, np.zeros_like(x_)), axis=-1)
    elif boundary == "periodic":
        pass
    else:
        raise ValueError("boundary must be one of 'mirror', 'zeros', or 'periodic'.")

    time_length = np.shape(x)[time_axis]
    time_length_ = np.shape(x_)[-1]

    # pad wavelet with zeros: JML ok
    order_length, freq_length, _ = np.shape(wavelet)
    _wavelet = np.zeros((order_length, freq_length, time_length_), dtype=np.cdouble)

    index = slice(
        int(np.floor(time_length_ - time_length) / 2),
        int(time_length + np.floor(time_length_ - time_length) / 2),
    )
    _wavelet[:, :, index] = wavelet_

    # take fft along axis = -1
    _wavelet_fft = np.fft.fft(_wavelet)
    om = 2 * np.pi * np.linspace(0, 1 - 1 / time_length_, time_length_)
    if time_length_ % 2 == 0:
        _wavelet_fft = (
            _wavelet_fft
            * np.exp(1j * -om * (time_length_ + 1) / 2)
            * np.sign(np.pi - om)
        )
    else:
        _wavelet_fft = _wavelet_fft * np.exp(1j * -om * (time_length_ + 1) / 2)

    # here we should be able to automate the tiling without assuming extra dimensions of wave
    X_ = np.tile(
        np.expand_dims(np.fft.fft(x_), (-3, -2)),
        (1, order_length, freq_length, 1),
    )

    # finally the transform; return precision of input `x``; central part only
    complex_dtype = np.cdouble if x.dtype == np.single else np.csingle
    wtx = np.fft.ifft(X_ * np.conj(_wavelet_fft)).astype(complex_dtype)
    wtx = wtx[..., index]

    # reposition the time axis if needed from axis -1
    if time_axis != -1:
        wtx = np.moveaxis(wtx, -1, time_axis)

    # remove extra dimensions if needed
    wtx = np.squeeze(wtx)

    return wtx


def morse_wavelet(
    length: int,
    gamma: float,
    beta: float,
    radian_frequency: np.ndarray,
    order: int = 1,
    normalization: str = "bandpass",
) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute the generalized Morse wavelets of Olhede and Walden (2002), doi: 10.1109/TSP.2002.804066.

    Parameters
    ----------
    length : int
       Length of the wavelets.
    gamma : float
       Gamma parameter of the wavelets.
    beta : float
       Beta parameter of the wavelets.
    radian_frequency : np.ndarray
       The radian frequencies at which the Fourier transform of the wavelets
       reach their maximum amplitudes. radian_frequency is between 0 and 2 * np.pi * 0.5,
       the normalized Nyquist radian frequency.
    order : int, optional
        Order of wavelets, default is 1.
    normalization : str, optional
       Normalization for the ``wavelet`` output. By default it is assumed to be ``"bandpass"``
       which uses a bandpass normalization, meaning that the FFT of the wavelets
       have peak value of 2 for all central frequencies ``radian_frequency``. The other option is
       ``"energy"``which uses the unit energy normalization. In this last case, the time-domain wavelet
       energies ``np.sum(np.abs(wave)**2)`` are always unity.

    Returns
    -------
    wavelet : np.ndarray
        Time-domain wavelets with shape (order, radian_frequency, length).
    wavelet_fft: np.ndarray
        Frequency-domain wavelets with shape (order, radian_frequency, length).

    Examples
    --------
    Compute a Morse wavelet with gamma parameter 3, beta parameter 4, at radian
    frequency 0.2 cycles per unit time:

    >>> wavelet, wavelet_fft = morse_wavelet(1024, 3, 4, np.array([2*np.pi*0.2]))
    >>> np.shape(wavelet)
    (1, 1, 1024)

    Compute a suite of Morse wavelets with gamma parameter 3, beta parameter 4, up to order 3,
    at radian frequencies 0.2 and 0.3 cycles per unit time:

    >>> wavelet, wavelet_fft = morse_wavelet(1024, 3, 4, np.array([2*np.pi*0.2, 2*np.pi*0.3]), order=3)
    >>> np.shape(wavelet)
    (3, 2, 1024)

    Compute a Morse wavelet specifying an energy normalization :
    >>> wavelet, wavelet_fft = morse_wavelet(1024, 3, 4, np.array([2*np.pi*0.2]), normalization="energy")

    Raises
    ------
    ValueError
        If normalization optional argument is not in ["bandpass", "energy"]``.

    See Also
    --------
    :func:`wavelet_transform`, :func:`morse_wavelet_transform`, :func:`morse_freq`, :func:`morse_logspace_freq`, :func:`morse_amplitude`, :func:`morse_properties`
    """
    # ad test for radian_frequency being a numpy array
    # initialization
    wavelet = np.zeros((length, order, len(radian_frequency)), dtype=np.cdouble)
    waveletfft = np.zeros((length, order, len(radian_frequency)), dtype=np.cdouble)

    # call to morse_wavelet take only gamma and be as float, no array
    fo, _, _ = morse_freq(gamma, beta)
    for i in range(len(radian_frequency)):
        wavelet_tmp = np.zeros((length, order), dtype=np.cdouble)
        waveletfft_tmp = np.zeros((length, order), dtype=np.cdouble)

        # wavelet frequencies
        fact = np.abs(radian_frequency[i]) / fo
        # norm_radian_frequency first dim is n points
        norm_radian_frequency = (
            2 * np.pi * np.linspace(0, 1 - 1 / length, length) / fact
        )
        if normalization == "energy":
            with np.errstate(divide="ignore"):
                waveletzero = np.exp(
                    beta * np.log(norm_radian_frequency) - norm_radian_frequency**gamma
                )
        elif normalization == "bandpass":
            if beta == 0:
                waveletzero = 2 * np.exp(-(norm_radian_frequency**gamma))
            else:
                with np.errstate(divide="ignore"):
                    waveletzero = 2 * np.exp(
                        -beta * np.log(fo)
                        + fo**gamma
                        + beta * np.log(norm_radian_frequency)
                        - norm_radian_frequency**gamma
                    )
        else:
            raise ValueError(
                "Normalization option (norm) must be one of 'energy' or 'bandpass'."
            )
        waveletzero[0] = 0.5 * waveletzero[0]
        # Replace NaN with zeros in waveletzero
        waveletzero = np.nan_to_num(waveletzero, copy=False, nan=0.0)
        # second family is never used
        waveletfft_tmp = _morse_wavelet_first_family(
            fact,
            gamma,
            beta,
            norm_radian_frequency,
            waveletzero,
            order=order,
            normalization=normalization,
        )
        waveletfft_tmp = np.nan_to_num(waveletfft_tmp, posinf=0, neginf=0)
        # shape of waveletfft_tmp is points, order
        # center wavelet
        norm_radian_frequency_mat = np.tile(
            np.expand_dims(norm_radian_frequency, -1), (order)
        )
        waveletfft_tmp = waveletfft_tmp * np.exp(
            1j * norm_radian_frequency_mat * (length + 1) / 2 * fact
        )
        # time domain waveletlet
        wavelet_tmp = np.fft.ifft(waveletfft_tmp, axis=0)
        if radian_frequency[i] < 0:
            wavelet[:, :, i] = np.conj(wavelet_tmp)
            waveletfft_tmp[1:-1, :] = np.flip(waveletfft_tmp[1:-1, :], axis=0)
            waveletfft[:, :, i] = waveletfft_tmp
        else:
            waveletfft[:, :, i] = waveletfft_tmp
            wavelet[:, :, i] = wavelet_tmp

    # reorder dimension to be (order, frequency, time steps)
    # enforce length 1 for first axis if order=1 (no squeezing)
    wavelet = np.moveaxis(wavelet, [0, 1, 2], [2, 0, 1])
    waveletfft = np.moveaxis(waveletfft, [0, 1, 2], [2, 0, 1])

    return wavelet, waveletfft


def _morse_wavelet_first_family(
    fact: float,
    gamma: float,
    beta: float,
    norm_radian_frequency: np.ndarray,
    wavezero: np.ndarray,
    order: int = 1,
    normalization: str = "bandpass",
) -> np.ndarray:
    """
    Derive first family of Morse wavelets. Internal use only.
    """
    r = (2 * beta + 1) / gamma
    c = r - 1
    L = np.zeros_like(norm_radian_frequency, dtype=np.float64)
    wavefft1 = np.zeros((np.shape(wavezero)[0], order))

    for i in np.arange(0, order):
        if normalization == "energy":
            A = morse_amplitude(gamma, beta, order=i + 1, normalization=normalization)
            coeff = np.sqrt(1 / fact) * A
        elif normalization == "bandpass":
            if beta != 0:
                coeff = np.sqrt(np.exp(_lgamma(r) + _lgamma(i + 1) - _lgamma(i + r)))
            else:
                coeff = 1

        index = slice(
            0, int(np.round(np.shape(wavezero)[0] / 2))
        )  # how to define indices?
        L[index] = _laguerre(2 * norm_radian_frequency[index] ** gamma, i, c)
        wavefft1[:, i] = coeff * wavezero * L

    return wavefft1


def morse_freq(
    gamma: np.ndarray | float,
    beta: np.ndarray | float,
) -> tuple[np.ndarray, np.ndarray, np.ndarray] | tuple[float, float, float]:
    """
    Frequency measures for generalized Morse wavelets. This functions calculates
    three different measures fm, fe, and fi of the frequency of the lowest-order generalized Morse
    wavelet specified by parameters ``gamma`` and ``beta``.

    Note that all frequency quantities here are in *radian* as in cos(f t) and not
    cyclic as in np.cos(2 np.pi f t).

    For ``beta=0``, the corresponding wavelet becomes an analytic lowpass filter, and fm
    is not defined in the usual way but as the point at which the filter has decayed
    to one-half of its peak power.

    For details see Lilly and Olhede (2009), doi: 10.1109/TSP.2008.2007607.

    Parameters
    ----------
    gamma : np.ndarray or float
       Gamma parameter of the wavelets.
    beta : np.ndarray or float
       Beta parameter of the wavelets.

    Returns
    -------
    fm : np.ndarray
        The modal or peak frequency.
    fe : np.ndarray
        The energy frequency.
    fi : np.ndarray
        The instantaneous frequency at the wavelets' centers.

    Examples
    --------
    >>> fm, fe, fi = morse_freq(3, 4)

    >>> morse_freq(3, 4)
    (array(1.10064242), 1.1025129235952809, 1.1077321674324723)

    >>> morse_freq(3, np.array([10, 20, 30]))
    (array([1.49380158, 1.88207206, 2.15443469]),
    array([1.49421505, 1.88220264, 2.15450116]),
    array([1.49543843, 1.88259299, 2.15470024]))

    >>> morse_freq(np.array([3, 4, 5]), np.array([10, 20, 30]))
    (array([1.49380158, 1.49534878, 1.43096908]),
    array([1.49421505, 1.49080278, 1.4262489 ]),
    array([1.49543843, 1.48652036, 1.42163583]))

    >>> morse_freq(np.array([3, 4, 5]), 10)
    (array([1.49380158, 1.25743343, 1.14869835]),
    array([1.49421505, 1.25000964, 1.13759731]),
    array([1.49543843, 1.24350315, 1.12739747]))

    See Also
    --------
    :func:`morse_wavelet`, :func:`morse_amplitude`
    """
    with np.errstate(divide="ignore"):  # ignore warning when beta=0
        fm = np.where(
            beta == 0,
            np.log(2) ** (1 / gamma),
            np.exp((1 / gamma) * (np.log(beta) - np.log(gamma))),
        )

    fe = (
        1
        / (2 ** (1 / gamma))
        * _gamma((2 * beta + 2) / gamma)
        / _gamma((2 * beta + 1) / gamma)
    )

    fi = _gamma((beta + 2) / gamma) / _gamma((beta + 1) / gamma)

    return fm, fe, fi


def morse_logspace_freq(
    gamma: float,
    beta: float,
    length: int,
    highset: tuple[float, float] = (0.1, np.pi),
    lowset: tuple[float, float] = (5, 0),
    density: int = 4,
) -> np.ndarray:
    """
    Compute logarithmically-spaced frequencies for generalized Morse wavelets
    with parameters gamma and beta. This is a useful function to obtain the frequencies
    needed for time-frequency analyses using wavelets. If ``radian_frequencies`` is the
    output, ``np.log(radian_frequencies)`` is uniformly spaced, following convention
    for wavelet analysis. See Lilly (2017), doi: 10.1098/rspa.2016.0776.

    Default settings to compute the frequencies can be changed by passing optional
    arguments ``lowset``, ``highset``, and ``density``. See below.

    Parameters
    ----------
    gamma : float
       Gamma parameter of the Morse wavelets.
    beta : float
       Beta parameter of the Morse wavelets.
    length : int
        Length of the Morse wavelets and input signals.
    highset : tuple of floats, optional.
        Tuple of values (eta, high) used for high-frequency cutoff calculation. The highest
        frequency is set to be the minimum of a specified value and a cutoff frequency
        based on a Nyquist overlap condition: the highest frequency is the minimum of
        the specified value high, and the largest frequency for which the wavelet will
        satisfy the threshold level eta. Here eta be a number between zero and one
        specifying the ratio of a frequency-domain wavelet at the Nyquist frequency
        to its peak value. Default is (eta, high) = (0.1, np.pi).
    lowset : tuple of floats, optional.
        Tupe of values (P, low) set used for low-frequency cutoff calculation based on an
        endpoint overlap condition. The lowest frequency is set such that the lowest-frequency
        wavelet will reach some number P, called the packing number, times its central window
        width at the ends of the time series. A choice of P=1 corresponds to  roughly 95% of
        the time-domain wavelet energy being contained within the time series endpoints for
        a wavelet at the center of the domain. The second value of the tuple is the absolute
        lowest frequency. Default is (P, low) = (5, 0).
    density : int, optional
        This optional argument controls the number of points in the returned frequency
        array. Higher values of ``density`` mean more overlap in the frequency
        domain between transforms. When ``density=1``, the peak of one wavelet is located at the
        half-power points of the adjacent wavelet. The default ``density=4`` means
        that four other wavelets will occur between the peak of one wavelet and
        its half-power point.

    Returns
    -------
    radian_frequency : np.ndarray
        Logarithmically-spaced frequencies in radians cycles per unit time,
        sorted in descending order.

    Examples
    --------
    Generate a frequency array for the generalized Morse wavelet
    with parameters gamma=3 and beta=5 for a time series of length n=1024:

    >>> radian_frequency = morse_logspace_freq(3, 5, 1024)
    >>> radian_frequency = morse_logspace_freq(3, 5, 1024, highset=(0.2, np.pi), lowset=(5, 0))
    >>> radian_frequency = morse_logspace_freq(3, 5, 1024, highset=(0.2, np.pi), lowset=(5, 0), density=10)

    See Also
    --------
    :func:`morse_wavelet`, :func:`morse_freq`, :func:`morse_properties`
    """
    gamma_ = np.array([gamma])
    beta_ = np.array([beta])
    width, _, _ = morse_properties(gamma_, beta_)

    _high = _morsehigh(gamma_, beta_, highset[0])
    high_ = np.min(np.append(_high, highset[1]))

    low = 2 * np.sqrt(2) * width * lowset[0] / length
    low_ = np.max(np.append(low, lowset[1]))

    r = 1 + 1 / (density * width)
    m = np.floor(np.log10(high_ / low_) / np.log10(r)).astype(int)[0]
    radian_frequency = high_ * np.ones(int(m + 1)) / r ** np.arange(0, m + 1)

    return radian_frequency


def _morsehigh(
    gamma: np.ndarray,
    beta: np.ndarray,
    eta: float,
) -> np.ndarray | float:
    """High-frequency cutoff of the generalized Morse wavelets.
    gamma and be should be arrays of the same length. Internal use only.
    """
    m = 10000
    omhigh = np.linspace(0, np.pi, m)
    f = np.zeros_like(gamma, dtype="float")

    for i in range(0, len(gamma)):
        fm, _, _ = morse_freq(gamma[i], beta[i])
        with np.errstate(all="ignore"):
            om = fm * np.pi / omhigh
            lnwave1 = beta[i] / gamma[i] * np.log(np.exp(1) * gamma[i] / beta[i])
            lnwave2 = beta[i] * np.log(om) - om ** gamma[i]
            lnwave = lnwave1 + lnwave2
        index = np.nonzero(np.log(eta) - lnwave < 0)[0][0]
        f[i] = omhigh[index]

    return f


def morse_properties(
    gamma: np.ndarray | float,
    beta: np.ndarray | float,
) -> tuple[np.ndarray, np.ndarray, np.ndarray] | tuple[float, float, float]:
    """
    Calculate the properties of the demodulated generalized Morse wavelets.
    See Lilly and Olhede (2009), doi: 10.1109/TSP.2008.2007607.

    Parameters
    ----------
    gamma : np.ndarray or float
       Gamma parameter of the wavelets.
    beta : np.ndarray or float
       Beta parameter of the wavelets.

    Returns
    -------
    width : np.ndarray or float
        Dimensionless time-domain window width of the wavelets.
    skew : np.ndarray or float
        Imaginary part of normalized third moment of the time-domain demodulate,
        or 'demodulate skewness'.
    kurt : np.ndarray or float
        Normalized fourth moment of the time-domain demodulate,
        or 'demodulate kurtosis'.

    Examples
    --------
    TODO

    See Also
    --------
    :func:`morse_wavelet`, :func:`morse_freq`, :func:`morse_amplitude`, :func:`morse_logspace_freq`.
    """
    # test common size? or could be broadcasted
    width = np.sqrt(gamma * beta)
    skew = (gamma - 3) / width
    kurt = 3 - skew**2 - 2 / width**2

    return width, skew, kurt


def morse_amplitude(
    gamma: np.ndarray | float,
    beta: np.ndarray | float,
    order: int = 1,
    normalization: str = "bandpass",
) -> float:
    """
    Calculate the amplitude coefficient of the generalized Morse wavelets.
    By default, the amplitude is calculated such that the maximum of the
    frequency-domain wavelet is equal to 2, which is the bandpass normalization.
    Optionally, specify ``normalization="energy"`` in order to return the coefficient
    giving the wavelets unit energies. See Lilly and Olhede (2009), doi doi: 10.1109/TSP.2008.2007607.

    Parameters
    ----------
    gamma : np.ndarray or float
       Gamma parameter of the wavelets.
    beta : np.ndarray or float
       Beta parameter of the wavelets.
    order : int, optional
        Order of wavelets, default is 1.
    normalization : str, optional
       Normalization for the wavelets. By default it is assumed to be ``"bandpass"``
       which uses a bandpass normalization, meaning that the FFT of the wavelets
       have peak value of 2 for all central frequencies ``radian_frequency``. The other option is ``"energy"``
       which uses the unit energy normalization. In this last case the time-domain wavelet
       energies ``np.sum(np.abs(wave)**2)`` are always unity.

    Returns
    -------
    amp : np.ndarray or float
        The amplitude coefficient of the wavelets.

    Examples
    --------
    TODO

    See Also
    --------
    :func:`morse_wavelet`, :func:`morse_freq`, :func:`morse_properties`, :func:`morse_logspace_freq`.
    """
    # add test for type and shape in case of ndarray
    if normalization == "energy":
        r = (2 * beta + 1) / gamma
        amp = (
            2 * np.pi * gamma * (2**r) * np.exp(_lgamma(order) - _lgamma(order + r - 1))
        ) ** 0.5
    elif normalization == "bandpass":
        fm, _, _ = morse_freq(gamma, beta)
        amp = np.where(beta == 0, 2, 2 / (np.exp(beta * np.log(fm) - fm**gamma)))
    else:
        raise ValueError(
            "Normalization option (normalization) must be one of 'energy' or 'bandpass'."
        )

    return amp


def _laguerre(
    x: np.ndarray | float,
    k: float,
    c: float,
) -> np.ndarray:
    """Generalized Laguerre polynomials"""
    y = np.zeros_like(x, dtype="float")
    for i in np.arange(0, k + 1):
        fact = np.exp(_lgamma(k + c + 1) - _lgamma(c + i + 1) - _lgamma(k - i + 1))
        y = y + (-1) ** i * fact * x**i / _gamma(i + 1)
    return y



import numpy as np

import tests.utils as testutils
from clouddrift import adapters, datasets
from clouddrift.ragged import apply_ragged, subset


class datasets_tests(testutils.DisableProgressTestCase):
    def test_gdp1h(self):
        with datasets.gdp1h() as ds:
            self.assertTrue(ds)

    def test_gdp6h(self):
        with datasets.gdp6h() as ds:
            self.assertTrue(ds)

    def test_ibtracs(self):
        """
        The original dataset contains data points for entire time span of the dataset. Due to this, we trim
        the original data array to only contain up to the numobs associated to the storm. In this test we validate
        the trimmed data and ensure the left over is all NaN.
        """
        options = dict(version="v04r01", kind="LAST_3_YEARS")
        ragged_ds = adapters.ibtracs.to_raggedarray(**options)
        ds = adapters.ibtracs._get_original_dataset(**options)
        ragged_ds_first = subset(
            ragged_ds, {"storm": 0}, row_dim_name="storm", rowsize_var_name="numobs"
        )
        ds_first = ds.sel(storm=0)
        first_numobs: np.ndarray = ds_first.numobs.data
        first_numobs = first_numobs.astype(int)

        # Both transformed and original data array should contain the same values
        self.assertTrue(
            np.all(
                np.allclose(
                    ds_first.usa_r34.data[:first_numobs],
                    ragged_ds_first.usa_r34.data,
                    equal_nan=True,
                )
            )
        )

        # The rest of the values should be nan
        self.assertTrue(
            np.all(
                np.isnan(
                    ds_first.usa_r34.data[first_numobs + 1 :],
                )
            )
        )

    def test_glad(self):
        with datasets.glad() as ds:
            self.assertTrue(ds)

    def test_glad_dims_coords(self):
        with datasets.glad() as ds:
            self.assertTrue(len(ds.sizes) == 2)
            self.assertTrue("obs" in ds.dims)
            self.assertTrue("traj" in ds.dims)
            self.assertTrue(len(ds.coords) == 2)
            self.assertTrue("time" in ds.coords)
            self.assertTrue("id" in ds.coords)

    def test_glad_subset_and_apply_ragged_work(self):
        with datasets.glad() as ds:
            ds_sub = subset(
                ds,
                {"id": ["CARTHE_001", "CARTHE_002"]},
                id_var_name="id",
                row_dim_name="traj",
            )
            self.assertTrue(ds_sub)
            mean_lon = apply_ragged(np.mean, [ds_sub.longitude], ds_sub.rowsize)
            self.assertTrue(mean_lon.size == 2)

    def test_spotters_opens(self):
        with datasets.spotters() as ds:
            self.assertTrue(ds)

    def test_subsurface_floats_opens(self):
        with datasets.subsurface_floats() as ds:
            self.assertTrue(ds)

    def test_andro_opens(self):
        with datasets.andro() as ds:
            self.assertTrue(ds is not None)
            self.assertTrue(len(ds.variables) > 0)
            self.assertTrue(len(ds["lon_d"]) > 0)

            self.assertTrue(
                len(ds.lat_d[np.logical_or(ds.lat_d > 90, ds.lat_d < -90)]) == 0
            )
            self.assertTrue(
                len(ds.lat_d[np.logical_or(ds.lon_d > 180, ds.lon_d < -180)]) == 0
            )

    def test_yomaha_opens(self):
        with datasets.yomaha() as ds:
            self.assertTrue(ds)

    def test_mosaic_opens(self):
        with datasets.mosaic() as ds:
            self.assertTrue(ds)



import os
import shutil
import unittest

import numpy as np

from clouddrift.adapters import gdp_source
from clouddrift.ragged import unpack


class gdp_source_files_integration(unittest.TestCase):
    def test_load_and_create_aggregate(self):
        """
        Test that each drifters segment has maintianed the row order as its read in
        from the data files. This is tested since the process of generating the
        aggregate dataset is performed in parallel per chunk (default 100_000 rows).

        Each chunk will contain a separate dataset per drifter in the chunk. Since two chunks
        may contain rows for one drifter (chunking can split a drifters trajectory) we join
        these partitioned segments in parallel per drifter.
        """
        ds = gdp_source.to_raggedarray(max=1, chunk_size=1_000, max_chunks=100)
        assert ds is not None

        all_drifter_obs_index = unpack(ds["obs_index"].data, ds["rowsize"])
        for drifter_obs_index in all_drifter_obs_index:
            obs_index_diffs = np.diff(drifter_obs_index, axis=0)
            self.assertEqual(
                np.all(obs_index_diffs == 1).T,
                True,
                "obs_index values not monotonically increasing",
            )

        start_dt_diffs = np.diff(ds["start_date"].data, axis=0)
        self.assertEqual(
            np.all(start_dt_diffs.astype(np.int64) >= 0).T,
            True,
            "Drifter segments not ordered by start date",
        )

        agg_path = os.path.join(gdp_source._TMP_PATH, "aggregate")
        os.makedirs(agg_path, exist_ok=True)
        ds.to_zarr(os.path.join(agg_path, "gdpsource_1f_agg.zarr"))

    @classmethod
    def tearDownClass(cls):
        [shutil.rmtree(dir) for dir in [gdp_source._TMP_PATH]]
